#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-13
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING, Type


if TYPE_CHECKING:
    from gort.gort import Gort, GortClient


__all__ = ["recipes", "BaseRecipe"]


recipes: dict[str, Type[BaseRecipe]] = {}


class RegisterRecipe(type):
    """Metaclass to register recipes."""

    def __new__(cls, name, bases, class_dict):
        cls = type.__new__(cls, name, bases, class_dict)

        if name == "BaseRecipe":
            return cls

        assert issubclass(cls, BaseRecipe)

        if "name" not in class_dict:
            raise ValueError(f"name attribute not defined in {name}.")

        # This is a hacked version of abc.ABCMeta, but this way we
        # don't need to chain two metaclasses.
        if "recipe" not in class_dict:
            raise ValueError(f"recipe method not defined in {name}.")

        recipes[class_dict["name"]] = cls

        return cls


class BaseRecipe(object, metaclass=RegisterRecipe):
    """Base class for recipes."""

    name: str | None = None

    def __init__(self, gort: Gort | GortClient):
        self.gort = gort

    async def recipe(self, *args, **kwargs):
        """The recipe. Must be overridden by the subclass."""

        return

    async def __call__(self, *args, **kwargs):
        await self.recipe(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-08
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import os

import click

from sdsstools.daemonizer import DaemonGroup, cli_coro


@click.group()
def gort():
    """Gort CLI."""

    pass


@gort.group(cls=DaemonGroup, prog="gort_ws", workdir=os.getcwd())
@cli_coro()
async def websocket():
    """Launches the websocket server."""

    from gort.websocket import WebsocketServer

    ws = WebsocketServer()
    await ws.start()

    await ws.websocket_server.serve_forever()


def main():
    gort()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-08
# @Filename: websocket.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
from functools import partial

from typing import TYPE_CHECKING, Callable, Coroutine

from websockets.legacy.protocol import broadcast
from websockets.server import WebSocketServerProtocol, serve

from gort import config
from gort.gort import Gort


if TYPE_CHECKING:
    from clu import AMQPReply


__all__ = ["WebsocketServer"]


CALLBACKS: dict[str, Callable] = {}
CB_TYPE = Callable[["WebsocketServer", WebSocketServerProtocol, str], None | Coroutine]


def route(name: str):
    """Defines a WebsocketServer router."""

    def decorator(fn: CB_TYPE):
        CALLBACKS[name] = fn
        return fn

    return decorator


class WebsocketServer:
    """A websocket server that allows communication with Gort.

    Parameters
    ----------
    whost
        The host where to run the websocket server. Defaults to
        ``config.websocket.host``.
    wport
        The TCP port on which to run the websocket server. Defaults to
        ``config.websocket.port``
    client_kwargs
        Arguments to pass to the `.Gort` client.

    """

    def __init__(
        self,
        whost: str | None = None,
        wport: int | None = None,
        **client_kwargs,
    ):
        self.gort = Gort(**client_kwargs)

        self.wparams = (
            whost or config["websocket"]["host"],
            wport or config["websocket"]["port"],
        )
        self.wclients: set[WebSocketServerProtocol] = set()

    async def start(self):
        """Start the server and AMQP client."""

        # self.gort.add_reply_callback(self._handle_reply)
        await self.gort.init()

        self.websocket_server = await serve(
            self._handle_websocket_connection,
            *self.wparams,
        )

        return self

    async def stop(self):
        """Stop the server and AMQP client."""

        await self.gort.stop()
        self.websocket_server.close()

    async def _handle_websocket_connection(self, websocket: WebSocketServerProtocol):
        """Handle a connection to the websocket server."""

        # Register the client
        self.wclients.add(websocket)

        async for data in websocket:
            try:
                message = json.loads(data)
                if not isinstance(message, dict):
                    continue
            except ValueError:
                continue

            route = message.get("route", "")
            command_id = message.get("command_id", "")
            params = message.get("params", {})

            if route in CALLBACKS:
                cb = CALLBACKS[route]
                if asyncio.iscoroutinefunction(cb):
                    asyncio.create_task(
                        cb(
                            self,
                            client=websocket,
                            command_id=command_id,
                            **params,
                        )
                    )
                else:
                    loop = asyncio.get_running_loop()
                    cb = partial(cb, self, command_id, **params)
                    loop.call_soon(cb)

        self.wclients.remove(websocket)

    async def _handle_reply(self, reply: AMQPReply):
        """Broadcast a reply to the connected websockets."""

        message = reply.message
        data = dict(
            headers=message.headers,
            exchange=message.exchange,
            message_id=message.message_id,
            routing_key=message.routing_key,
            timestamp=message.timestamp.isoformat() if message.timestamp else None,
            body=reply.body,
        )
        broadcast(self.wclients, json.dumps(data))

    async def reply_to_client(
        self,
        client: WebSocketServerProtocol,
        command_id: str,
        message: dict = {},
        **message_kwargs,
    ):
        """Reply to a client command.

        Parameters
        ----------
        client
            The client websocket connection to use to reply to the client.
        command_id
            The command ID associated with the client command.
        message
            The message to send to the client.
        message_kwargs
            Keyword arguments that will be used to update the message,

        """

        final_message = {"command_id": command_id}
        final_message.update(message)
        final_message.update(message_kwargs)

        try:
            await client.send(json.dumps(final_message))
        except Exception as err:
            self.gort.log.warning(f"Failed replying to WS client: {err}.")

    @route("enclosure_status")
    async def enclosure_status(
        self,
        client: WebSocketServerProtocol,
        command_id: str,
        **_,
    ):
        """Returns the enclosure status."""

        status = await self.gort.enclosure.status()
        await self.reply_to_client(client, command_id, status)

    @route("enclosure_action")
    async def enclosure_action(
        self,
        client: WebSocketServerProtocol,
        command_id: str,
        action: str | None = None,
    ):
        """Opens/closes/stops the enclosure."""

        if action == "open":
            await self.gort.enclosure.open()
            await self.reply_to_client(client, command_id, {"text": "Dome open"})
        elif action == "close":
            await self.gort.enclosure.close()
            await self.reply_to_client(client, command_id, {"text": "Dome closed"})
        elif action == "stop":
            await self.gort.enclosure.stop()
            await self.reply_to_client(client, command_id, {"text": "Dome stopped"})
        else:
            await self.reply_to_client(client, command_id, {"error": "Invalid action"})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-09
# @Filename: observer.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
import re
from time import time

from typing import TYPE_CHECKING, Any

import numpy
import pandas
from astropy.time import Time

from gort.exceptions import GortObserverError
from gort.exposure import Exposure
from gort.tile import Coordinates
from gort.tools import build_guider_reply_list, cancel_task


if TYPE_CHECKING:
    from gort.gort import Gort
    from gort.tile import Tile


__all__ = ["GortObserver"]


class GortObserver:
    """A class to handle tile observations.

    Parameters
    ----------
    gort
        The instance of :obj:`.Gort` used to communicate with the devices.
    tile
        The :obj:`.Tile` with the information about the observation.
    mask_positions_pattern
        The ``spec`` fibre mask positions to use.

    """

    def __init__(self, gort: Gort, tile: Tile, mask_positions_pattern: str = "P1-*"):
        self.gort = gort
        self.tile = tile

        self.mask_positions = self._get_mask_positions(mask_positions_pattern)

        self.guide_task: asyncio.Future | None = None
        self.guider_monitor = GuiderMonitor(self)

        self.standards = Standards(self, tile, self.mask_positions)

        self.__current_exposure: Exposure | None = None

    def __repr__(self):
        return f"<GortObserver (tile_id={self.tile.tile_id})>"

    @property
    def has_standards(self):
        """Returns :obj:`True` if standards will be observed."""

        return len(self.standards.standards) > 0

    async def slew(self):
        """Slew to the telescope fields."""

        tile = self.tile

        cotasks = []

        # Stops guiders.
        await self.gort.guiders.stop()

        # Slew telescopes.
        self.write_to_log(f"Slewing to tile_id={tile.tile_id}.", level="info")

        sci = (tile.sci_coords.ra, tile.sci_coords.dec, tile.sci_coords.pa)
        self.write_to_log(f"Science: {str(tile.sci_coords)}")

        spec = None
        if tile.spec_coords and len(tile.spec_coords) > 0:
            first_spec = tile.spec_coords[0]

            # For spec we slew to the fibre with which we'll observe first.
            # This should save a bit of time converging.
            spec_target = (first_spec.ra, first_spec.dec)
            # spec = fibre_slew_coordinates(*spec_target, self.mask_positions[0])
            spec = spec_target

            self.write_to_log(f"Spec: {first_spec} on {self.mask_positions[0]}")

        sky = {}
        for skytel in ["SkyE", "SkyW"]:
            if skytel.lower() in tile.sky_coords:
                sky_coords_tel = tile.sky_coords[skytel.lower()]
                if sky_coords_tel is not None:
                    sky[skytel.lower()] = (sky_coords_tel.ra, sky_coords_tel.dec)
                    self.write_to_log(f"{skytel}: {sky_coords_tel}")

        cotasks.append(
            self.gort.telescopes.goto(
                sci=sci,
                spec=spec,
                skye=sky.get("skye", None),
                skyw=sky.get("skyw", None),
            )
        )

        # Move fibsel to first position.
        fibsel = self.gort.telescopes.spec.fibsel
        cotasks.append(fibsel.move_to_position(self.mask_positions[0]))

        # Execute.
        await asyncio.gather(*cotasks)

    async def acquire(self, guide_tolerance: float = 3, timeout: float = 180):
        """Acquires the field in all the telescopes. Blocks until then.

        Parameters
        ----------
        guide_tolerance
            The guide tolerance in arcsec. A telescope will not be considered
            to be guiding if its separation to the commanded field is larger
            than this value.
        timeout
            The maximum time allowed for acquisition. In case of timeout
            the acquired fields are evaluated and an exception is
            raised if the acquisition failed.

        Raises
        ------
        GortObserverError
            If the acquisition failed or the minimum required telescopes are
            not guiding.

        """

        # Determine telescopes on which to guide.

        guide_coros = []
        guide_on_telescopes: list[str] = []
        n_skies = 0
        for tel in ["sci", "skye", "skyw", "spec"]:
            coords = self.tile[tel]

            if coords is None or (isinstance(coords, list) and len(coords) == 0):
                continue

            if tel == "spec" and isinstance(coords, list):
                coords = coords[0]
            assert isinstance(coords, Coordinates)

            guide_on_telescopes.append(tel)
            if "sky" in tel:
                n_skies += 1

            # Pixel in the MF on which to guide. Always None/central pixel
            # except for sci if defined. For spec we guide on the pixel of the
            # first fibre/mask position.
            pixel = coords._mf_pixel if tel != "spec" else self.mask_positions[0]

            guide_coros.append(
                self.gort.guiders[tel].guide(
                    ra=coords.ra,
                    dec=coords.dec,
                    guide_tolerance=guide_tolerance,
                    pixel=pixel,
                )
            )

        if "spec" not in guide_on_telescopes:
            self.write_to_log("No standards defined. Blocking fibre mask.", "warning")
            await self.gort.telescopes.spec.fibsel.move_relative(500)

        if n_skies == 0:
            self.write_to_log("No sky positions defined.", "warning")

        # Start guide loop.
        await self.guider_monitor.restart()
        self.guide_task = asyncio.gather(*guide_coros)

        await asyncio.sleep(5)

        # Wait until convergence.
        self.write_to_log("Waiting for guiders to converge.")
        guide_status = await asyncio.gather(
            *[
                self.gort.guiders[tel].wait_until_guiding(timeout=timeout)
                for tel in guide_on_telescopes
            ]
        )

        has_timedout = any([gs[3] for gs in guide_status])
        if has_timedout:
            self.write_to_log("Some acquisitions timed out.", "warning")
            # TODO: we need to stop guiders that timed out.

        try:
            for ii, tel in enumerate(guide_on_telescopes):
                is_guiding = guide_status[ii][0]
                if tel == "sci" and not is_guiding:
                    raise GortObserverError(
                        "Science telescope is not guiding.",
                        error_code=801,
                    )
                if tel == "spec" and not is_guiding:
                    raise GortObserverError(
                        "Spec telescope is not guiding.",
                        error_code=801,
                    )
                if "sky" in tel and not is_guiding:
                    self.write_to_log(f"{tel} telescope is not guiding.", "warning")

        except Exception:
            self.write_to_log("Stopping guide loops.", "warning")
            await self.gort.guiders.stop()

            raise

        self.write_to_log("All telescopes are now guiding.")

    async def expose(
        self,
        exposure_time: float = 900.0,
        show_progress: bool | None = None,
        count: int = 1,
        object: str | None = None,
    ):
        """Starts exposing the spectrographs.

        Parameters
        ----------
        exposure_time
            The length of the exposure in seconds.
        show_progress
            Displays a progress bar with the elapsed exposure time.
        count
            Number of exposures. If ``iterate_over_standards=True``, a
            full sequence of standards will be observed during each
            exposure.
        object
            The object name to be added to the header.

        Returns
        -------
        exposures
            Either a single `.Exposure` or a list of exposures if ``count>1``.

        """

        tile_id = self.tile.tile_id
        dither_pos = self.tile.dither_position

        if object is None:
            object = self.tile.object

        exposures: list[Exposure] = []

        for nexp in range(1, count + 1):
            self.write_to_log(
                f"Starting {exposure_time:.1f} s exposure ({nexp}/{count}).",
                "info",
            )

            # Refresh guider data for this exposure.
            if nexp > 1:
                await self.guider_monitor.restart()

            # Move fibre selector to the first position. Should be there unless
            # count > 1 in which case we need to move it back.
            await self.standards.start_iterating(exposure_time)

            exposure = Exposure(self.gort, flavour="object", object=object)
            self.__current_exposure = exposure

            exposure.hooks["pre-readout"].append(self._pre_readout)

            await exposure.expose(
                exposure_time=exposure_time,
                show_progress=show_progress,
            )

            exposures.append(exposure)

            await exposure.register_observation(tile_id=tile_id, dither_pos=dither_pos)

        if len(exposures) == 1:
            return exposures[0]
        else:
            return exposures

    async def finish_observation(self):
        """Finishes the observation, stops the guiders, etc."""

        self.write_to_log("Finishing observation.", "info")

        # Should have been cancelled in _update_header(), but just in case.
        await self.standards.cancel()

        if self.guide_task is not None and not self.guide_task.done():
            await self.gort.guiders.stop()
            await self.guide_task

    def write_to_log(
        self,
        message: str,
        level: str = "debug",
        header: str | None = None,
    ):
        """Writes a message to the log with a custom header.

        Parameters
        ----------
        message
            The message to log.
        level
            The level to use for logging: ``'debug'``, ``'info'``, ``'warning'``, or
            ``'error'``.
        header
            The header to prepend to the message. By default uses the class name.

        """

        if header is None:
            header = f"({self.__class__.__name__}) "

        message = f"{header}{message}"

        level = logging.getLevelName(level.upper())
        assert isinstance(level, int)

        self.gort.log.log(level, message)

    def _get_mask_positions(self, pattern: str):
        """Returns mask positions sorted by motor steps."""

        mask_config = self.gort.config["telescopes"]["mask_positions"]
        all_positions = self.gort.telescopes.spec.fibsel.list_positions()
        positions = [pos for pos in all_positions if re.match(pattern, pos)]

        return sorted(positions, key=lambda p: mask_config[p])

    async def _pre_readout(self, header: dict[str, Any]):
        """Updates the exposure header with pointing and guiding information."""

        tile_id = self.tile.tile_id
        dither_pos = self.tile.dither_position

        tile_header = {
            "TILE_ID": (tile_id or -999, "The tile_id of this observation"),
            "DPOS": (dither_pos, "Dither position"),
            "POSCIRA": round(self.tile.sci_coords.ra, 6),
            "POSCIDE": round(self.tile.sci_coords.dec, 6),
        }

        if self.tile["skye"]:
            tile_header.update(
                {
                    "POSKYERA": round(self.tile.sky_coords["skye"].ra, 6),
                    "POSKYEDE": round(self.tile.sky_coords["skye"].dec, 6),
                }
            )

        if self.tile["skyw"]:
            tile_header.update(
                {
                    "POSKYWRA": round(self.tile.sky_coords["skyw"].ra, 6),
                    "POSKYWDE": round(self.tile.sky_coords["skyw"].dec, 6),
                }
            )

        header.update(tile_header)

        self.guider_monitor.update_data()
        header.update(self.guider_monitor.to_header())

        # At this point the shutter is closed so let's stop observing standards.
        # This also finishes updating the standards table.
        if self.has_standards:
            await self.standards.cancel()

            # There is some overhead between when we set t0 for the first standard
            # and when the exposure actually begins. This leads to the first standard
            # having longer exposure time than open shutter.
            if self.has_standards and self.__current_exposure is not None:
                start_time = self.__current_exposure.start_time.unix
                self.standards.standards.loc[1, "t0"] = start_time

            header.update(self.standards.to_header())


class GuiderMonitor:
    """Monitors guider exposures."""

    def __init__(self, observer: GortObserver):
        self.observer = observer
        self.gort = observer.gort

        self.guider_task: asyncio.Task | None = None
        self.guider_data: pandas.DataFrame | None = None

        self._current_data = []

    async def start(self):
        """Starts monitoring."""

        self.guider_task = asyncio.create_task(self._guider_monitor())

    async def stop(self):
        """Stops monitoring."""

        await cancel_task(self.guider_task)

    async def restart(self):
        """Restarts monitoring."""

        self.guider_data = None

        await self.stop()
        await self.start()

    async def _guider_monitor(self):
        """Monitors the guider data and build a data frame."""

        self._current_data = []

        task = asyncio.create_task(
            build_guider_reply_list(
                self.gort,
                self._current_data,
            )
        )

        try:
            while True:
                # Just keep the task running.
                await asyncio.sleep(1)

        except asyncio.CancelledError:
            await cancel_task(task)
            self.update_data()

    def update_data(self):
        """Updates the guider data frame."""

        if len(self._current_data) > 0:
            # Build DF with all the frames.
            df = pandas.DataFrame.from_records(self._current_data)

            # Group by frameno, keep only non-NaN values.
            df = df.groupby(["frameno", "telescope"], as_index=False).apply(
                lambda g: g.fillna(method="bfill", axis=0).iloc[0, :]
            )

            # Remove NaN rows.
            df = df.dropna()

            # Sort by frameno.
            df = df.sort_values("frameno")

            self.guider_data = df

    def to_header(self):
        """Returns a header with pointing and guiding information."""

        header: dict[str, Any] = {}

        if self.guider_data is not None:
            for tel in ["sci", "spec", "skye", "skyw"]:
                try:
                    tel_data = self.guider_data.loc[self.guider_data.telescope == tel]
                    if len(tel_data) < 2:
                        frame0 = None
                        framen = None
                    else:
                        frame0 = int(tel_data.frameno.min())
                        framen = int(tel_data.frameno.max())

                    header.update(
                        {
                            f"G{tel.upper()}FR0": (frame0, f"{tel} first guider frame"),
                            f"G{tel.upper()}FRN": (framen, f"{tel} last guider frame"),
                        }
                    )

                except Exception as err:
                    self.gort.specs.write_to_log(
                        f"Failed updating guider header information for {tel}: {err}",
                        "warning",
                    )
                    continue

        return header


class Standards:
    """Iterates over standards and monitors observed standards."""

    def __init__(self, observer: GortObserver, tile: Tile, mask_positions: list[str]):
        self.observer = observer
        self.gort = observer.gort

        self.tile = tile

        self.iterate_task: asyncio.Task | None = None

        self.current_standard: int = 1
        self.standards = self._get_frame()

    @property
    def mask_positions(self):
        """Returns the list of mask positions from `.GortObserver`."""

        return self.observer.mask_positions

    def _get_frame(self):
        """Constructs the standard data frame."""

        stdn = list(range(1, len(self.tile.spec_coords) + 1))
        source_id = [cc.source_id or -1 for cc in self.tile.spec_coords]
        ra = [cc.ra for cc in self.tile.spec_coords]
        dec = [cc.dec for cc in self.tile.spec_coords]

        default = [0] * len(stdn)

        df = pandas.DataFrame(
            {
                "n": pandas.Series(stdn, dtype=numpy.int16),
                "source_id": pandas.Series(source_id, dtype=numpy.int64),
                "ra": pandas.Series(ra, dtype=numpy.float64),
                "dec": pandas.Series(dec, dtype=numpy.float64),
                "acquired": pandas.Series(default, dtype=numpy.int16),
                "observed": pandas.Series(default, dtype=numpy.int16),
                "t0": pandas.Series(default, dtype=numpy.float64),
                "t1": pandas.Series(default, dtype=numpy.float64),
                "fibre": pandas.Series([""] * len(stdn), dtype=str),
            }
        )

        df.set_index("n", inplace=True)

        return df

    async def start_iterating(self, exposure_time: float):
        """Iterates over the fibre mask positions.

        Moving the ``spec`` telescope to each one of the standard star,
        acquires the new field, and adjusts the time to stay on each target.

        """

        if len(self.standards) == 0:
            return

        await self.gort.telescopes.spec.fibsel.move_to_position(self.mask_positions[0])
        await self.cancel()

        self.standards = self._get_frame()
        self.current_standard = 1

        self.standards.loc[1, "acquired"] = 1
        self.standards.loc[1, "t0"] = time()
        self.standards.loc[1, "fibre"] = self.mask_positions[0]

        self.iterate_task = asyncio.create_task(self._iterate(exposure_time))

    async def cancel(self):
        """Cancels iteration."""

        await cancel_task(self.iterate_task)

        if len(self.standards) == 0:
            return

        if self.standards.loc[self.current_standard].acquired == 1:
            if self.standards.loc[self.current_standard, "observed"] != 1:
                self.standards.loc[self.current_standard, "observed"] = 1
                self.standards.loc[self.current_standard, "t1"] = time()

    async def _iterate(self, exposure_time: float):
        """Iterate task."""

        # Time to acquire a standard.
        ACQ_PER_STD = 30

        spec_coords = self.tile.spec_coords

        # If we have zero or one standards, do nothing. The spec telescope
        # is already pointing to the first mask position.
        if len(spec_coords) <= 1:
            return

        guider_pixels: dict[str, tuple[float, float]]
        guider_pixels = self.gort.config["guiders"]["devices"]["spec"]["named_pixels"]

        # Time at which the exposure began.
        t0 = time()

        # Time to spend on each mask position.
        n_stds = len(spec_coords)

        # Calculate the time to actually be on target, taking into account
        # how long we expect to take acquiring.
        time_per_position = exposure_time / n_stds - ACQ_PER_STD
        if time_per_position < 0:
            self.observer.write_to_log(
                "Exposure time is too short to observe this "
                "many standards. I will do what I can.",
                "warning",
            )
            time_per_position = exposure_time / n_stds

        # Time at which we started observing the last standard.
        t0_last_std = t0

        # Number of standards observed.
        n_observed = 1

        # Index of current standard being observed.
        current_std_idx = 0

        while True:
            await asyncio.sleep(1)

            # We consider than if there is less 2 * ACQ_PER_STD left in the
            # exposure there is no point in going to the next standard.
            t_now = time()
            if t_now - t0 > exposure_time - 2 * ACQ_PER_STD:
                self.observer.write_to_log("Exiting standard loop.")
                self.observer.write_to_log(
                    f"Standards observed: {n_observed}/{n_stds}.",
                    "info",
                )
                return

            # Time to move to another standard?
            if t_now - t0_last_std > time_per_position:
                # Check that we haven't run out standards. If so,
                # keep observing this one.
                if len(spec_coords) == current_std_idx + 1:
                    continue

                # Moving the mask to an intermediate position while we move around.
                spec_tel = self.gort.telescopes.spec
                await spec_tel.fibsel.move_relative(500)

                # Register the previous standard.
                self.standards.loc[self.current_standard, "t1"] = time()
                self.standards.loc[self.current_standard, "observed"] = True

                # Increase current index and get coordinates.
                current_std_idx += 1
                self.current_standard += 1

                # New coordinates to observe.
                new_coords = spec_coords[current_std_idx]
                new_mask_position = self.mask_positions[current_std_idx]

                # Pixel on the MF corresponding to the new fibre/mask hole on
                # which to guide. We use this tabulated list instead of
                # offset_to_master_frame_pixel() because the latter coordinates
                # are less precise as they do not include IFU rotation and more
                # precise metrology.
                new_guider_pixel = guider_pixels[new_mask_position]

                self.observer.write_to_log(
                    f"Moving to standard #{current_std_idx+1} ({new_coords}) "
                    f"on fibre {new_mask_position}.",
                    "info",
                )

                # Slew to new coordinates. We actually slew to the coordinates
                # that make the new star close to the fibre that will observe it.
                # slew_ra, slew_dec = fibre_slew_coordinates(
                #     new_coords.ra,
                #     new_coords.dec,
                #     new_mask_position,
                # )
                slew_ra = new_coords.ra
                slew_dec = new_coords.dec

                # Finish guiding on spec telescope.
                self.observer.write_to_log("Stopping guiding on spec and reslewing.")

                cotasks = [
                    self.gort.guiders["spec"].stop(),
                    spec_tel.goto_coordinates(ra=slew_ra, dec=slew_dec),
                ]
                await asyncio.gather(*cotasks)

                # Start to guide. Note that here we use the original coordinates
                # of the star along with the pixel on the master frame on which to
                # guide. See the note in fibre_slew_coordinates().
                self.observer.write_to_log("Starting to guide on spec telescope.")
                asyncio.create_task(
                    self.gort.guiders.spec.guide(
                        ra=new_coords.ra,
                        dec=new_coords.dec,
                        guide_tolerance=3,
                        pixel=new_guider_pixel,
                    )
                )

                result = await self.gort.guiders.spec.wait_until_guiding(timeout=60)
                if result[0] is False:
                    self.observer.write_to_log(
                        "Failed to acquire standard position. Skipping.",
                        "warning",
                    )
                    continue

                self.observer.write_to_log(f"Standard {new_mask_position} acquired.")

                # Move mask to uncover fibre.
                await spec_tel.fibsel.move_to_position(new_mask_position)

                n_observed += 1
                t0_last_std = time()

                self.standards.loc[self.current_standard, "acquired"] = 1
                self.standards.loc[self.current_standard, "t0"] = time()
                self.standards.loc[self.current_standard, "fibre"] = new_mask_position

    def to_header(self):
        """Returns observed standards as a header-ready dictionary."""

        header_data = {}

        for nstd, data in self.standards.iterrows():
            header_data[f"STD{nstd}ID"] = data.source_id if data.source_id > 0 else None
            header_data[f"STD{nstd}RA"] = data.ra
            header_data[f"STD{nstd}DE"] = data.dec
            header_data[f"STD{nstd}ACQ"] = bool(data.observed)

            if data.observed:
                header_data[f"STD{nstd}T0"] = Time(data.t0, format="unix").isot
                header_data[f"STD{nstd}T1"] = Time(data.t1, format="unix").isot
                header_data[f"STD{nstd}EXP"] = round(data.t1 - data.t0, 1)
                header_data[f"STD{nstd}FIB"] = data.fibre

        return header_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-10
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import concurrent.futures
import pathlib
import re
import tempfile
import warnings
from contextlib import suppress
from datetime import datetime
from functools import partial

from typing import TYPE_CHECKING, Callable, Coroutine

import httpx
import pandas
import peewee
from astropy import units as uu
from astropy.coordinates import angular_separation as astropy_angular_separation

from gort import config


if TYPE_CHECKING:
    from pyds9 import DS9

    from clu import AMQPClient, AMQPReply

    from gort.devices.telescope import FibSel
    from gort.gort import GortClient


__all__ = [
    "get_valid_variable_name",
    "ds9_agcam_monitor",
    "parse_agcam_filename",
    "ds9_display_frames",
    "get_next_tile_id",
    "get_calibrators",
    "get_next_tile_id_sync",
    "get_calibrators_sync",
    "register_observation",
    "get_ccd_frame_path",
    "move_mask_interval",
    "angular_separation",
    "get_db_connection",
    "run_in_executor",
    "is_interactive",
    "is_notebook",
    "cancel_task",
    "build_guider_reply_list",
    "get_temporary_file_path",
]

CAMERAS = [
    "sci.west",
    "sci.east",
    "skye.west",
    "skye.east",
    "skyw.west",
    "skyw.east",
    "spec.east",
]


def get_valid_variable_name(var_name: str):
    """Converts a string to a valid variable name."""

    return re.sub(r"\W|^(?=\d)", "_", var_name)


async def ds9_agcam_monitor(
    amqp_client: AMQPClient,
    cameras: list[str] | None = None,
    replace_path_prefix: tuple[str, str] | None = None,
    **kwargs,
):
    """Shows guider images in DS9."""

    images_handled = set([])

    # Clear all frames and get an instance of DS9.
    ds9 = await ds9_display_frames([], clear_frames=True, preserve_frames=False)

    if cameras is None:
        cameras = CAMERAS.copy()

    agcam_actors = set(
        [
            "lvm." + (cam.split(".")[0] if "." in cam else cam) + ".agcam"
            for cam in cameras
        ]
    )

    async def handle_reply(reply: AMQPReply):
        sender = reply.sender
        if sender not in agcam_actors:
            return

        message: dict | None = None
        if "east" in reply.body:
            message = reply.body["east"]
        elif "west" in reply.body:
            message = reply.body["west"]
        else:
            return

        if message is None or message.get("state", None) != "written":
            return

        filename: str = message["filename"]
        if filename in images_handled:
            return
        images_handled.add(filename)

        if replace_path_prefix is not None:
            filename = filename.replace(replace_path_prefix[0], replace_path_prefix[1])

        await ds9_display_frames([filename], ds9=ds9, **kwargs)

    amqp_client.add_reply_callback(handle_reply)

    while True:
        await asyncio.sleep(1)


async def ds9_display_frames(
    files: list[str | pathlib.Path] | dict[str, str | pathlib.Path],
    ds9: DS9 | None = None,
    order=CAMERAS,
    ds9_target: str = "DS9:*",
    show_all_frames=True,
    preserve_frames=True,
    clear_frames=False,
    adjust_zoom=True,
    adjust_scale=True,
    show_tiles=True,
):
    """Displays a series of images in DS9."""

    if ds9 is None:
        try:
            import pyds9
        except ImportError:
            raise ImportError("pyds9 is not installed.")

        ds9 = pyds9.DS9(target=ds9_target)

    if clear_frames:
        ds9.set("frame delete all")

    files_dict: dict[str, str] = {}
    if not isinstance(files, dict):
        for file_ in files:
            tel_cam = parse_agcam_filename(file_)
            if tel_cam is None:
                raise ValueError(f"Cannot parse type of file {file_!s}.")
            files_dict[".".join(tel_cam)] = str(file_)
    else:
        files_dict = {k: str(v) for k, v in files.items()}

    nframe = 1
    for cam in order:
        if cam in files_dict:
            file_ = files_dict[cam]
            ds9.set(f"frame {nframe}")
            ds9.set(f"fits {file_}")
            if adjust_scale:
                ds9.set("zscale")
            if adjust_zoom:
                ds9.set("zoom to fit")
            nframe += 1
        else:
            if show_all_frames:
                if preserve_frames is False:
                    ds9.set(f"frame {nframe}")
                    ds9.set("frame clear")
                nframe += 1

    if show_tiles:
        ds9.set("tile")

    return ds9


def parse_agcam_filename(file_: str | pathlib.Path) -> tuple[str, str] | None:
    """Returns the type of an ``agcam`` file in the form ``(telescope, camera)``."""

    file_ = pathlib.Path(file_)
    basename = file_.name

    match = re.match(".+(sci|spec|skyw|skye).+(east|west)", basename)
    if not match:
        return None

    return match.groups()


def get_next_tile_id_sync() -> dict:
    """Retrieves the next ``tile_id`` from the scheduler API. Synchronous version."""

    sch_config = config["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    with httpx.Client(base_url=f"http://{host}:{port}/") as client:
        resp = client.get("next_tile")
        if resp.status_code != 200:
            raise httpx.RequestError("Failed request to /next_tile")
        tile_id_data = resp.json()

    return tile_id_data


async def get_next_tile_id() -> dict:
    """Retrieves the next ``tile_id`` from the scheduler API."""

    sch_config = config["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    async with httpx.AsyncClient(base_url=f"http://{host}:{port}/") as client:
        resp = await client.get("next_tile")
        if resp.status_code != 200:
            raise httpx.RequestError("Failed request to /next_tile")
        tile_id_data = resp.json()

    return tile_id_data


def get_calibrators_sync(
    tile_id: int | None = None,
    ra: float | None = None,
    dec: float | None = None,
) -> dict:
    """Get calibrators for a ``tile_id`` or science pointing. Synchronous version."""

    sch_config = config["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    with httpx.Client(base_url=f"http://{host}:{port}/") as client:
        if tile_id:
            resp = client.get("cals", params={"tile_id": tile_id})
        elif ra is not None and dec is not None:
            resp = client.get("cals", params={"ra": ra, "dec": dec})
        else:
            raise ValueError("ra and dec are required.")
        if resp.status_code != 200:
            raise httpx.RequestError("Failed request to /cals")

    return resp.json()


async def get_calibrators(
    tile_id: int | None = None,
    ra: float | None = None,
    dec: float | None = None,
):
    """Get calibrators for a ``tile_id`` or science pointing."""

    sch_config = config["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    async with httpx.AsyncClient(base_url=f"http://{host}:{port}/") as client:
        if tile_id:
            resp = await client.get("cals", params={"tile_id": tile_id})
        elif ra is not None and dec is not None:
            resp = await client.get("cals", params={"ra": ra, "dec": dec})
        else:
            raise ValueError("ra and dec are required.")
        if resp.status_code != 200:
            raise httpx.RequestError("Failed request to /cals")

    return resp.json()


async def register_observation(payload: dict):
    """Registers an observation with the scheduler."""

    sch_config = config["scheduler"]
    host = sch_config["host"]
    port = sch_config["port"]

    async with httpx.AsyncClient() as client:
        resp = await client.put(
            f"http://{host}:{port}/register_observation",
            json=payload,
            follow_redirects=True,
        )

        if resp.status_code != 200 or not resp.json()["success"]:
            raise RuntimeError(f"Failed registering observation: {resp.text}.")


def is_notebook() -> bool:
    """Returns :obj:`True` if the code is run inside a Jupyter Notebook.

    https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook

    """

    try:
        shell = get_ipython().__class__.__name__  # type: ignore
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def is_interactive():
    """Returns :obj:`True` is we are in an interactive session."""

    import __main__ as main

    return not hasattr(main, "__file__")


def get_ccd_frame_path(
    frame_id: int,
    sjd: int | None = None,
    cameras: str | list[str] | None = None,
    spectro_path="/data/spectro",
) -> list[str]:
    """Returns the paths for the files for a spectrograph frame.

    Parameters
    ----------
    frame_id
        The spectrograph frame for which the paths are searched.
    mjd
        The SJD in which the frames where taken. If not provided, all the
        directories under ``spectro_path`` are searched.
    cameras
        The cameras to be returned. If :obj:`None`, all cameras found are returned.
    spectro_path
        The path to the ``spectro`` directory where spectrograph files are
        stored under an SJD structure.

    Returns
    -------
    paths
        The list of paths to CCD frames that match ``frame_id``.

    """

    if isinstance(cameras, str):
        cameras = [cameras]

    base_path = pathlib.Path(spectro_path)
    recursive = True
    if sjd:
        base_path /= str(sjd)
        recursive = False

    # Glob all files that match the frame_id.
    globp = f"*{frame_id}.fits.*"
    if recursive:
        globp = f"**/{globp}"

    files = [str(path) for path in base_path.glob(globp)]

    if cameras is None:
        return files

    files_camera = []
    for camera in cameras:
        for file in files:
            if f"-{camera}-" in file:
                files_camera.append(file)

    return files_camera


async def move_mask_interval(
    gort: GortClient,
    positions: str | list[str] = "P1-*",
    order_by_steps: bool = False,
    total_time: float | None = None,
    time_per_position: float | None = None,
    notifier: Callable[[str], None] | Callable[[str], Coroutine] | None = None,
):
    """Moves the fibre mask in the spectrophotometric telescope at intervals.

    Parameters
    ----------
    gort
        The instance of :obj:`.Gort` to communicate with the actor system.
    positions
        The positions to iterate over. It can be a string in which case it will
        be treated as a regular expression and any mask position that matches the
        value will be iterated, in alphabetic order. Alternative it can be a list
        of positions to move to which will be executed in that order.
    order_by_steps
        If :obj:`True`, the positions are iterated in order of smaller to larger
        number of step motors.
    total_time
        The total time to spend iterating over positions, in seconds. Each position
        will  be visited for an equal amount of time. The time required to move the
        mask will not be taken into account, which means the total execution
        time will be longer than ``total_time``.
    time_per_position
        The time to spend on each mask position, in seconds. The total execution
        time will be ``len(positions)*total_time+overhead`` where ``overhead`` is
        the time required to move the mask between positions.
    notifier
        A function or coroutine to call every time a new position is reached.
        If it's a coroutine, it is scheduled as a task. If it is a normal
        callback it should run quickly to not perceptibly affect the total
        execution time.

    """

    try:
        fibsel: FibSel = gort.telescopes.spec.fibsel
    except Exception as err:
        raise RuntimeError(f"Cannot find fibre selector: {err}")

    if total_time is not None and time_per_position is not None:
        raise ValueError("Only one of total_time or time_per_position can be used.")

    if total_time is None and time_per_position is None:
        raise ValueError("One of total_time or time_per_position needs to be passed.")

    mask_config = gort.config["telescopes"]["mask_positions"]
    all_positions = list(mask_config)

    if isinstance(positions, str):
        regex = positions
        all_positions = fibsel.list_positions()
        positions = [pos for pos in all_positions if re.match(regex, pos)]

        if order_by_steps:
            positions = sorted(positions, key=lambda p: mask_config[p])

    fibsel.write_to_log(f"Iterating over positions {positions}.")

    if total_time:
        time_per_position = total_time / len(positions)

    assert time_per_position is not None

    for position in positions:
        await fibsel.move_to_position(position)

        # Notify.
        if notifier is not None:
            if asyncio.iscoroutinefunction(notifier):
                asyncio.create_task(notifier(position))
            else:
                notifier(position)

        await asyncio.sleep(time_per_position)


def angular_separation(lon1: float, lat1: float, lon2: float, lat2: float):
    """A wrapper around astropy's ``angular_separation``.

    Returns the separation between two sets of coordinates. All units must
    be degrees and the returned values is also the separation in degrees.

    """

    separation = astropy_angular_separation(
        lon1 * uu.degree,  # type: ignore
        lat1 * uu.degree,  # type: ignore
        lon2 * uu.degree,  # type: ignore
        lat2 * uu.degree,  # type: ignore
    )

    return separation.to("deg").value


def get_db_connection():
    """Returns a DB connection from the configuration file parameters."""

    conn = peewee.PostgresqlDatabase(**config["database"])
    assert conn.connect(), "Database connection failed."

    return conn


async def run_in_executor(fn, *args, catch_warnings=False, executor="thread", **kwargs):
    """Runs a function in an executor.

    In addition to streamlining the use of the executor, this function
    catches any warning issued during the execution and reissues them
    after the executor is done. This is important when using the
    actor log handler since inside the executor there is no loop that
    CLU can use to output the warnings.

    In general, note that the function must not try to do anything with
    the actor since they run on different loops.

    """

    fn = partial(fn, *args, **kwargs)

    if executor == "thread":
        executor = concurrent.futures.ThreadPoolExecutor
    elif executor == "process":
        executor = concurrent.futures.ProcessPoolExecutor
    else:
        raise ValueError("Invalid executor name.")

    if catch_warnings:
        with warnings.catch_warnings(record=True) as records:
            with executor() as pool:
                result = await asyncio.get_event_loop().run_in_executor(pool, fn)

        for ww in records:
            warnings.warn(ww.message, ww.category)

    else:
        with executor() as pool:
            result = await asyncio.get_running_loop().run_in_executor(pool, fn)

    return result


async def cancel_task(task: asyncio.Task | None):
    """Safely cancels a task."""

    if task is None or task.done():
        return

    task.cancel()
    with suppress(asyncio.CancelledError):
        await task


async def build_guider_reply_list(
    gort: GortClient,
    reply_list: list[dict],
    actor: str | None = None,
):
    """Tasks that monitors the guider output and builds a list of replies.

    This coroutine is meant to be run as a task. When the task is cancelled it
    will clean itself by removing the callback in the client.

    Parameters
    ----------
    gort
        The Gort client to connect to the actor system.
    reply_list
        A list (usually empty) to which the task will append the replies.
    actor
        The actor to listen to. If :obj:`None`, listens to all the guider actors.

    """

    async def handle_guider_reply(reply: AMQPReply):
        if actor is not None:
            if actor not in str(reply.sender):
                return
        else:
            if ".guider" not in str(reply.sender):
                return

        body = reply.body
        telescope = str(reply.sender).split(".")[1]

        if "frame" in body:
            frame = body["frame"]
            reply_list.append(
                {
                    "frameno": frame["seqno"],
                    "time": pandas.to_datetime(datetime.now()),
                    "n_sources": frame["n_sources"],
                    "focus_position": frame["focus_position"],
                    "fwhm": frame["fwhm"],
                    "telescope": telescope,
                }
            )
        elif "measured_pointing" in body:
            measured_pointing = body["measured_pointing"]
            reply_list.append(
                {
                    "frameno": measured_pointing["frameno"],
                    "ra": measured_pointing["ra"],
                    "dec": measured_pointing["dec"],
                    "ra_offset": measured_pointing["radec_offset"][0],
                    "dec_offset": measured_pointing["radec_offset"][1],
                    "separation": measured_pointing["separation"],
                    "mode": measured_pointing["mode"],
                    "telescope": telescope,
                }
            )
        elif "correction_applied" in body:
            correction_applied = body["correction_applied"]

            reply_list.append(
                {
                    "frameno": correction_applied["frameno"],
                    "ax0_applied": correction_applied["motax_applied"][0],
                    "ax1_applied": correction_applied["motax_applied"][1],
                    "telescope": telescope,
                }
            )
        else:
            return

    try:
        gort.add_reply_callback(handle_guider_reply)

        # If the list does not expand every 30s, clean and exit.
        while True:
            nlist = len(reply_list)
            await asyncio.sleep(30)
            if len(reply_list) == nlist:
                return

    finally:
        gort.remove_reply_callback(handle_guider_reply)


def get_temporary_file_path(*args, create_parents: bool = False, **kwargs):
    """Returns a valid path to a temporary file.

    `args` and `kwargs` are directly passed to `tempfile.NamedTemporaryFile`.
    If `create_parents`, the parent directories are created if they don't
    exist.

    """

    tmp_log_file = tempfile.NamedTemporaryFile(*args, **kwargs)
    tmp_log_file.close()

    tmp_path = pathlib.Path(tmp_log_file.name)
    if tmp_path.exists():
        tmp_path.unlink()
    elif create_parents:
        tmp_path.parent.mkdir(parents=True, exist_ok=True)

    return tmp_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-03-13
# @Filename: nps.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from gort.gort import GortDevice, GortDeviceSet


if TYPE_CHECKING:
    from gort import ActorReply
    from gort.gort import GortClient


__all__ = ["NPS", "NPSSet"]


class NPS(GortDevice):
    """Class representing a networked power switch."""

    def __init__(self, gort: GortClient, name: str, actor: str, **kwargs):
        super().__init__(gort, name, actor)

    async def status(self):
        """Retrieves the status of the power outlet."""

        reply: ActorReply = await self.actor.commands.status()
        return reply.flatten()["status"][self.name]

    async def on(self, outlet: str):
        """Turns an outlet on."""

        await self.actor.commands.on(outlet)

    async def off(self, outlet: str):
        """Turns an outlet on."""

        await self.actor.commands.off(outlet)

    async def all_off(self):
        """Turns off all the outlets."""

        for outlet_number in range(1, 9):
            self.write_to_log(f"Turning off outlet {outlet_number}.")
            await self.actor.commands.off("", outlet_number)


class NPSSet(GortDeviceSet[NPS]):
    """A set of networked power switches."""

    __DEVICE_CLASS__ = NPS
    __DEPLOYMENTS__ = ["lvmnps"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-02-07
# @Filename: core.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from types import SimpleNamespace

from typing import TYPE_CHECKING, Callable, Literal

import unclick
from aiormq import AMQPConnectionError, ChannelInvalidStateError
from typing_extensions import Self

from clu.tools import CommandStatus

from gort.exceptions import GortTimeoutError, GortWarning, RemoteCommandError

from .tools import get_valid_variable_name


if TYPE_CHECKING:
    from clu.client import AMQPReply
    from clu.command import Command

    from gort.gort import GortDevice

    from .gort import GortClient


__all__ = ["RemoteActor", "RemoteCommand", "ActorReply"]


class CommandSet(dict[str, "RemoteCommand"]):
    """A command set for a remote actor."""

    def __getattribute__(self, __name: str) -> RemoteCommand:
        if __name in self:
            return self[__name]
        return super().__getattribute__(__name)


class RemoteActor:
    """A programmatic representation of a remote actor."""

    def __init__(self, client: GortClient, name: str, device: GortDevice | None = None):
        self.client = client

        self.name = name
        self.model: dict = {}
        self.commands = CommandSet()
        self.device = device

    def __repr__(self):
        return f"<RemoteActor (name={self.name})>"

    async def init(self) -> Self:
        """Initialises the representation of the actor."""

        if (
            self.client.connection.connection
            and self.client.connection.connection.is_closed
        ):
            raise RuntimeError("Gort is not connected.")

        cmd = await self.send_raw_command("get-command-model")
        if cmd.status.did_fail:
            warnings.warn(f"Cannot get model for actor {self.name}.", GortWarning)
            return self

        self.model = cmd.replies.get("command_model")

        commands_dict = {}
        for command_info in self.model["commands"].values():
            command_name = get_valid_variable_name(command_info["name"])
            commands_dict[command_name] = RemoteCommand(self, command_info)

        self.commands = CommandSet(commands_dict)

        return self

    async def send_raw_command(self, *args, **kwargs):
        """Sends a raw command to the actor.

        The parameters are the same as CLU's ``AMQPClient.send_command()`` with
        the exception of the consumer name, which is replaced with the current actor
        name.

        """

        # This lock prevents that if the client has disconnected we get multiple
        # reconnections.
        async with self.client._connect_lock:
            try:
                cmd = await self.client.send_command(
                    self.name,
                    *args,
                    await_command=False,
                    **kwargs,
                )
            except (AMQPConnectionError, ChannelInvalidStateError):
                # Client has disconnected. This should only happen if running Gort
                # in an ipython terminal where the event loop only runs while a command
                # is executing. See https://tinyurl.com/4kcwxzx9

                self.client.log.warning("Client has disconnected. Reconnecting.")

                await self.client.start()

                cmd = await self.client.send_command(
                    self.name,
                    *args,
                    await_command=False,
                    **kwargs,
                )

        return await cmd

    async def refresh(self):
        """Refresesh the command list."""

        await self.init()


class RemoteCommand:
    """Representation of a remote command."""

    def __init__(
        self,
        remote_actor: RemoteActor,
        model: dict,
        parent: RemoteCommand | None = None,
    ):
        self._remote_actor = remote_actor
        self._model = model
        self._parent = parent

        self._name = model["name"]
        self.commands = SimpleNamespace()

        self.is_group = "commands" in model and len(model["commands"]) > 0
        if self.is_group:
            for command_info in model["commands"].values():
                command_name = get_valid_variable_name(command_info["name"])
                child_command = RemoteCommand(remote_actor, command_info, parent=self)
                setattr(self.commands, command_name, child_command)

    def get_command_string(self, *args, **kwargs):
        """Gets the command string for a set of arguments."""

        return unclick.build_command_string(self._model, *args, **kwargs)

    async def __call__(
        self,
        *args,
        reply_callback: Callable[[AMQPReply], None] | None | Literal[False] = None,
        timeout: float | None = None,
        **kwargs,
    ):
        """Executes the remote command with some given arguments."""

        parent_string = ""
        if self._parent is not None:
            # Call parent chain without arguments. This is not bullet-proof for all
            # cases, but probably good enough for now.
            parent_string = self._parent.get_command_string() + " "

        if reply_callback is None and self._remote_actor.device is not None:
            reply_callback = self._remote_actor.device.log_replies

        cmd = await self._remote_actor.send_raw_command(
            parent_string + self.get_command_string(*args, **kwargs),
            callback=reply_callback,
            time_limit=timeout,
        )

        actor_reply = ActorReply(self._remote_actor, cmd)
        for reply in cmd.replies:
            if len(reply.body) > 0:
                actor_reply.replies.append(reply.body)

        actor = self._remote_actor.name
        command_name = self._name

        if cmd.status & CommandStatus.TIMEDOUT:
            raise GortTimeoutError(
                f"Actor {actor!r} timed out executing command {command_name!r}.",
                command=cmd,
                remote_command=self,
            )

        if not cmd.status.did_succeed:
            error = actor_reply.get("error")
            error = f" {error!s}" if error is not None else ""
            raise RemoteCommandError(
                f"Actor {actor!r} failed executing command {command_name!r}.{error}",
                command=cmd,
                remote_command=self,
            )

        return actor_reply


@dataclass
class ActorReply:
    """A reply to an actor command.

    Attributes
    ----------
    actor
        The :obj:`.RemoteActor` that generated the reply.
    command
        The CLU command instance.
    replies
        A list of replies generated by the command.

    """

    actor: RemoteActor
    command: Command
    replies: list[dict] = field(default_factory=list)

    def flatten(self):
        """Returns a flattened dictionary of replies.

        Note that if a keyword has been output multiple times, only the
        last value is retained.

        """

        result = {}
        for reply in self.replies:
            for key in reply:
                result[key] = reply[key]

        return result

    def get(self, key: str):
        """Returns the first occurrence of a keyword in the reply list."""

        for reply in self.replies:
            if key in reply:
                return reply[key]

        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-25
# @Filename: ag.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from gort.gort import GortDevice, GortDeviceSet


__all__ = ["AG", "AGSet"]


class AG(GortDevice):
    """Class representing an AG camera."""

    async def status(self):
        """Returns the status of the AG."""

        return await self.actor.commands.status()

    async def reconnect(self):
        """Reconnect the AG cameras."""

        self.write_to_log("Reconnecting AG cameras.")

        return await self.actor.commands.reconnect()

    async def expose(
        self,
        exposure_time: float = 5.0,
        flavour: str = "object",
        **kwargs,
    ):
        """Exposes the cameras.

        Parameters
        ----------
        exposure_time
            The amount of time, in seconds, to expose the cameras.
        flavour
            The type of image to take, one of ``'bias'``, ``'dark'``, ``'flat'``,
            or ``'object'``.
        kwargs
            Any parameters to send to the ``lvmcam expose`` command.

        """

        if flavour == "bias":
            kwargs["bias"] = True
        elif flavour == "flat":
            kwargs["flat"] = True
        elif flavour == "dark":
            kwargs["dark"] = True
        else:
            kwargs["object"] = True

        return await self.actor.commands.expose(
            exptime=exposure_time,
            **kwargs,
        )


class AGSet(GortDeviceSet[AG]):
    """A set of auto-guiders."""

    __DEVICE_CLASS__ = AG
    __DEPLOYMENTS__ = ["lvmagcam"]

    async def reconnect(self):
        """Reconnects all the AG cameras.

        Returns
        -------
        cameras
            A list of cameras available after reconnecting.

        """

        await asyncio.gather(*[ag.reconnect() for ag in self.values()])

        return await self.list_alive_cameras()

    async def list_alive_cameras(self):
        """Returns a list of cameras found alive and well.

        Currently when a camera has disconnected, the status command still reports
        it as "online", but it doesn't report its temperature or other parameters.

        """

        all_status = await asyncio.gather(*[ag.status() for ag in self.values()])

        cameras = []
        for reply in all_status:
            actor = reply.actor.name
            telescope = actor.split(".")[1]
            messages = reply.flatten()
            if "east" in messages and "temperature" in messages["east"]:
                cameras.append(f"{telescope}-e")
            if "west" in messages and "temperature" in messages["west"]:
                cameras.append(f"{telescope}-w")

        return cameras

    async def expose(
        self,
        exposure_time: float = 5.0,
        flavour: str = "object",
        **kwargs,
    ):
        """Exposes the cameras.

        Parameters
        ----------
        exposure_time
            The amount of time, in seconds, to expose the cameras.
        flavour
            The type of image to take, one of ``'bias'``, ``'dark'``, ``'flat'``,
            or ``'object'``.
        kwargs
            Any parameters to send to the ``lvmcam expose`` command.

        """

        return await asyncio.gather(
            *[
                ag.expose(
                    exposure_time=exposure_time,
                    flavour=flavour,
                    **kwargs,
                )
                for ag in self.values()
            ]
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-29
# @Filename: kubernetes.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import datetime
import logging
import os
import pathlib
from time import sleep

import kubernetes
from kubernetes.utils import create_from_yaml

from gort import config
from gort.tools import is_notebook


class Kubernetes:
    """Interface with the Kubernetes cluster."""

    def __init__(self, log: logging.Logger | None = None):
        self.is_notebook = is_notebook()

        self.log = log or logging.getLogger("gort.kubernetes")

        if os.getenv("KUBERNETES_SERVICE_HOST"):
            self.is_pod = True
        else:
            self.is_pod = False

        # If we are in a notebook, we assume it's the one running in the Jupyter
        # Lab deployment, which is configured to have access to the cluster.
        if self.is_notebook or self.is_pod:
            kubernetes.config.load_incluster_config()
        else:
            kubernetes.config.load_config()

        self.v1 = kubernetes.client.CoreV1Api()
        self.apps_v1 = kubernetes.client.AppsV1Api()

    def list_namespaces(self):
        """Returns a list of namespaces."""

        namespace_info = self.v1.list_namespace()
        namespaces = [item.metadata.name for item in namespace_info.items]

        return namespaces

    def list_deployments(self):
        """Returns a list of deployments."""

        deployment_info = self.apps_v1.list_deployment_for_all_namespaces()
        deployments = [item.metadata.name for item in deployment_info.items]

        return deployments

    def get_deployment_namespace(self, deployment: str):
        """Returns the namespace of a deployment."""

        deployment_info = self.apps_v1.list_deployment_for_all_namespaces()

        for item in deployment_info.items:
            meta = item.metadata
            if meta.name == deployment:
                return meta.namespace

        return None

    def get_yaml_file(self, name: str):
        """Finds and returns the contents of a Kubernetes YAML file."""

        if self.is_notebook or self.is_pod:
            path = pathlib.Path(config["kubernetes"]["path"]["notebook"])
        else:
            path = pathlib.Path(config["kubernetes"]["path"]["default"])

        path = path.expanduser()

        files = list(path.glob(f"**/{name}*"))

        if files is None or len(files) == 0:
            raise ValueError(f"No YAML file found for {name!r}.")
        elif len(files) > 1:
            raise ValueError(f"Multiple YAML files found for {name!r}.")

        return files[0]

    def apply_from_file(self, name: str | pathlib.Path):
        """Applies a YAML file.

        Parameters
        ----------
        name
            The full path to the file to apply. If the path is relative,
            the file will be searched in the directory for YAML files
            defined in the configuration.

        """

        if isinstance(name, pathlib.Path) or os.path.isabs(name):
            path = pathlib.Path(name)
        else:
            path = self.get_yaml_file(name)

        deployments = create_from_yaml(
            kubernetes.client.ApiClient(),
            yaml_file=str(path),
        )

        return [dep[0].metadata.name for dep in deployments]

    def delete_deployment(self, deployment: str):
        """Deletes resources from a YAML file.

        Parameters
        ----------
        deployment
            The deployment to delete.

        """

        namespace = self.get_deployment_namespace(deployment)
        if namespace is None:
            raise ValueError(f"Deployment {deployment!r} not found.")

        self.apps_v1.delete_namespaced_deployment(deployment, namespace)

    def restart_deployment(self, deployment: str, from_file: bool = True):
        """Restarts a deployment.

        If the deployment is running, does a rollout restart. Otherwise looks
        for the deployment file and applies it.

        """

        if deployment in self.list_deployments() and not from_file:
            namespace = self.get_deployment_namespace(deployment)
            if namespace is None:
                raise ValueError(f"Namespace not found for deployment {deployment}.")

            # Create a patch for the current deployment saying that it was restarted
            # now, and it will.
            now = datetime.datetime.utcnow()
            now = str(now.isoformat("T") + "Z")
            body = {
                "spec": {
                    "template": {
                        "metadata": {
                            "annotations": {"kubectl.kubernetes.io/restartedAt": now}
                        }
                    }
                }
            }

            self.log.debug(f"Rollout restarting deployment {deployment}.")
            self.apps_v1.patch_namespaced_deployment(
                deployment,
                namespace,
                body,
                pretty="true",
            )

        else:
            try:
                file_ = self.get_yaml_file(deployment)
            except ValueError as err:
                self.log.warning(
                    f"Failed restarting from file: {err} "
                    "Doing rollout restart instead."
                )
                return self.restart_deployment(deployment, from_file=False)

            if deployment in self.list_deployments():
                self.delete_deployment(deployment)
                sleep(3)  # Give some time for the pods to exit.
            else:
                self.log.warning(f"{deployment!r} is not running.")

            self.log.info(f"Starting deployment from YAML file {str(file_)}.")
            self.apply_from_file(file_)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-18
# @Filename: enclosure.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING

from gort.exceptions import GortEnclosureError
from gort.gort import GortDevice, GortDeviceSet


if TYPE_CHECKING:
    from gort import ActorReply
    from gort.gort import GortClient


__all__ = ["Enclosure"]


class Enclosure(GortDevice):
    """Class representing the LVM enclosure."""

    __DEPLOYMENTS__ = ["lvmecp"]

    def __init__(self, gort: GortClient, name: str, actor: str, **kwargs):
        super().__init__(gort, name, actor)

    async def restart(self):
        """Restarts the ``lvmecp`` deployment."""

        # HACK: GortDevice has the everything that is needed to run restart.
        await GortDeviceSet.restart(self)  # type: ignore

    async def status(self):
        """Retrieves the status of the power outlet."""

        reply: ActorReply = await self.actor.commands.status()

        return reply.flatten()

    async def _prepare_telescopes(self):
        """Moves telescopes to park position before opening/closing the enclosure."""

        telescopes = list(self.gort.telescopes)
        is_parked = await asyncio.gather(
            *[self.gort.telescopes[tel].is_parked() for tel in telescopes]
        )

        if all(is_parked):
            return True

        self.write_to_log(
            "Moving telescopes to park before operating the dome.",
            "warning",
        )

        park_coros = [
            self.gort.telescopes[tel].goto_named_position("park")
            for itel, tel in enumerate(telescopes)
            if not is_parked[itel]
        ]

        await asyncio.gather(*park_coros)

    async def open(self, park_telescopes: bool = True):
        """Open the enclosure dome.

        Parameters
        ----------
        park_telescopes
            Move the telescopes to the park position before opening the
            enclosure to prevent dust or other debris falling on them.

        """

        if park_telescopes:
            await self._prepare_telescopes()

        self.write_to_log("Opening the enclosure ...", level="info")
        await self.actor.commands.dome.commands.open()
        self.write_to_log("Enclosure is now open.", level="info")

    async def close(self, park_telescopes: bool = True, force: bool = False):
        """Close the enclosure dome.

        Parameters
        ----------
        park_telescopes
            Move the telescopes to the park position before closing the
            enclosure to prevent dust or other debris falling on them.
        force
            Tries to closes the dome even if the system believes it is
            already closed.

        """

        if park_telescopes:
            try:
                await self._prepare_telescopes()
            except Exception as err:
                self.write_to_log(
                    f"Failed determining the status of the telescopes: {err}",
                    "warning",
                )
                if force is False:
                    raise GortEnclosureError(
                        "Not closing without knowing where the telescopes are. "
                        "If you really need to close call again with force=True."
                    )
                else:
                    self.write_to_log("Closing anyway because force=True", "warning")

        self.write_to_log("Closing the enclosure ...", level="info")
        await self.actor.commands.dome.commands.close(force=force)
        self.write_to_log("Enclosure is now closed.", level="info")

    async def stop(self):
        """Stop the enclosure dome."""

        self.write_to_log("Stoping the dome.", level="info")
        await self.actor.commands.dome.commands.stop()

    async def is_local(self):
        """Returns :obj:`True` if the enclosure is in local mode."""

        status = await self.status()
        safety_status_labels = status.get("safety_status_labels", None)
        if safety_status_labels is None:
            raise GortEnclosureError(
                "Cannot determine if enclosure is in local mode.",
                error_code=501,
            )

        # This should generally not be on, but it's useful as a way of disabling
        # the local mode when the lock or door are not working.
        if self.gort.config["enclosure"].get("bypass_local_mode", False) is True:
            return False

        return "LOCAL" in safety_status_labels

    async def get_door_status(self):
        """Returns the status of the door and lock."""

        status = await self.status()
        safety_status_labels = status.get("safety_status_labels", None)
        if safety_status_labels is None:
            raise GortEnclosureError(
                "Cannot determine door status.",
                error_code=502,
            )

        reply = {
            "door_closed": "DOOR_CLOSED" in safety_status_labels,
            "door_locked": "DOOR_LOCKED" in safety_status_labels,
            "local": await self.is_local(),
        }

        return reply



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-11
# @Filename: exposure.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import pathlib
import warnings
from collections import defaultdict

from typing import TYPE_CHECKING, Any, Awaitable, Callable

from astropy.io import fits
from astropy.time import Time
from rich.progress import BarColumn, MofNCompleteColumn, Progress, TextColumn

from sdsstools.time import get_sjd

from gort.exceptions import ErrorCodes, GortSpecError
from gort.tools import cancel_task, is_interactive, is_notebook, register_observation


if TYPE_CHECKING:
    from gort.gort import Gort, GortClient


__all__ = ["Exposure", "READOUT_TIME"]


READOUT_TIME = 51

HOOKS_TYPE = defaultdict[str, list[Callable[[Any], Awaitable]]]


class Exposure(asyncio.Future["Exposure"]):
    """A class representing an exposure taken by a :obj:`.SpectrographSet`.

    Parameters
    ----------
    gort
        A `.Gort` instance to communicate with the actors.
    exp_no
        The exposure sequence number. If :obj:`None`, the next valid
        sequence number will be used.
    flavour
        The image type. Defaults to ``'object'``.
    object
        The object name to be added to the header.

    Attributes
    ----------
    hooks
        A dictionary of hooks to call in specific steps of the exposure. Each
        hook must be a list of coroutines to call in that specific situation.
        All coroutines for a given hook are called concurrently and depending
        on the hook they may be scheduled as a task and not awaited. The only
        hook at this time is ``'pre-readout'`` which is called with the header
        before readout begins; the coroutine can modify the header in place or
        perform any tasks that should be complete at the end of integration.

        To add a coroutine to a hook ::

            async def update_header(header):
                header.update({'KEY': 1})

            exp = Exposure(g)
            exp.hooks['pre-readout'].append(update_header)

    """

    def __init__(
        self,
        gort: Gort | GortClient,
        exp_no: int | None = None,
        flavour: str | None = "object",
        object: str | None = "",
    ):
        self.specs = gort.specs
        self.exp_no = exp_no or self.specs.get_expno()
        self.flavour = flavour or "object"
        self.object = object or ""
        self.start_time = Time.now()

        self.error: bool = False
        self.reading: bool = False

        self._timer_task: asyncio.Task | None = None
        self._progress: Progress | None = None

        self.hooks: HOOKS_TYPE = defaultdict(
            list,
            {"pre-readout": []},
        )

        if self.flavour not in ["arc", "object", "flat", "bias", "dark"]:
            raise GortSpecError(
                "Invalid flavour type.",
                error_code=ErrorCodes.USAGE_ERROR,
            )

        super().__init__()

    def __repr__(self):
        return (
            f"<Exposure (exp_no={self.exp_no}, flavour={self.flavour}, "
            f"object={self.object!r} error={self.error}, reading={self.reading}, "
            f"done={self.done()})>"
        )

    async def expose(
        self,
        exposure_time: float | None = None,
        header: dict | None = None,
        async_readout: bool = False,
        show_progress: bool | None = None,
        object: str | None = None,
    ):
        """Exposes the spectrograph.

        Parameters
        ----------
        exposure_time
            The exposure time.
        header
            A dictionary with the extra header values..
        async_readout
            Returns after integration completes. Readout is initiated
            but handled asynchronously and can be await by awaiting
            the returned :obj:`.Exposure` object.
        show_progress
            Displays a progress bar with the elapsed exposure time.
            If :obj:`None` (the default), will show the progress bar only
            in interactive sessions.
        object
            The object name to be passed to the header.

        """

        # Check that all specs are idle and not errored.
        status = await self.specs.status(simple=True)
        for spec_status in status.values():
            if "IDLE" not in spec_status["status_names"]:
                raise GortSpecError(
                    "Some spectrographs are not IDLE.",
                    error_code=ErrorCodes.SECTROGRAPH_NOT_IDLE,
                )
            if "ERROR" in spec_status["status_names"]:
                raise GortSpecError(
                    "Some spectrographs have ERROR status. "
                    "Solve this manually before exposing.",
                    error_code=ErrorCodes.SECTROGRAPH_NOT_IDLE,
                )

        await self.specs.reset()

        header = header or {}

        # Set object name for header.
        if object is not None:
            header.update({"OBJECT": object})
        elif self.object is not None and self.object != "":
            header.update({"OBJECT": self.object})
        elif self.flavour != "object":
            header.update({"OBJECT": self.flavour})
        else:
            header.update({"OBJECT": ""})

        if show_progress is None:
            show_progress = is_interactive() or is_notebook()

        if exposure_time is None and self.flavour != "bias":
            raise GortSpecError(
                "Exposure time required for all flavours except bias.",
                error_code=ErrorCodes.USAGE_ERROR,
            )

        warnings.filterwarnings("ignore", message=".*cannot modify a done command.*")

        self.specs.last_exposure = self

        monitor_task: asyncio.Task | None = None

        try:
            if show_progress:
                await self.start_timer(exposure_time or 0.0)

            self.start_time = Time.now()

            await self.specs._send_command_all(
                "expose",
                exposure_time=exposure_time,
                seqno=self.exp_no,
                readout=False,
            )

            # Call pre-readout tasks.
            await self._call_hook("pre-readout", header)

            # At this point we have integrated and are ready to read.
            self.reading = True
            readout_task = asyncio.create_task(
                self.specs._send_command_all(
                    "read",
                    header=json.dumps(header),
                )
            )

            # Now launch the task that marks the Future done when the spec
            # is IDLE. If async_readout=False then that will return immediately
            # because the spec is already idle. If async_readout=True, this method
            # will return now and the task will mark the Future done when readout
            # complete (readout is ongoing and does not need to be launched).
            monitor_task = asyncio.create_task(self._done_monitor())
            if not async_readout:
                await readout_task
                await monitor_task
                self.specs.write_to_log(f"Exposure {self.exp_no} completed.")
            else:
                await self.stop_timer()
                self.specs.write_to_log("Returning with async readout ongoing.")

        except Exception as err:
            # Cancel the monitor task
            await cancel_task(monitor_task)

            self.error = True
            self.set_result(self)

            raise GortSpecError(f"Exposure failed with error {err}", error_code=301)

        finally:
            await self.stop_timer()

            if self.done() and not self.error:
                self.verify_files()

        return self

    async def start_timer(
        self,
        exposure_time: float,
        readout_time: float = READOUT_TIME,
    ):
        """Starts the rich timer."""

        self._progress = Progress(
            TextColumn(f"[yellow]({self.exp_no})"),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(bar_width=None),
            MofNCompleteColumn(),
            TextColumn("s"),
            expand=True,
            transient=True,
            auto_refresh=True,
            console=self.specs.gort._console,  # Need to use same console as logger.
        )

        exp_task = self._progress.add_task(
            "[blue] Integrating ...",
            total=int(exposure_time),
        )

        readout_task = self._progress.add_task(
            "[blue] Reading ...",
            total=int(readout_time),
            visible=False,
        )

        self._progress.start()

        async def update_timer():
            elapsed = 0
            while True:
                if elapsed > exposure_time + readout_time:
                    break
                elif self._progress is None:
                    return
                elif elapsed < exposure_time:
                    self._progress.update(exp_task, advance=1)
                else:
                    self._progress.update(
                        exp_task,
                        description="[green] Integration complete",
                        completed=int(exposure_time),
                    )
                    self._progress.update(readout_task, advance=1, visible=True)

                await asyncio.sleep(1)
                elapsed += 1

            if self._progress and readout_task:
                self._progress.update(
                    readout_task,
                    completed=int(readout_time),
                    description="[green] Readout complete",
                )

        def done_timer(*_):
            if self._progress:
                self._progress.stop()
                self._progress.console.clear_live()
                self._progress = None

        self._timer_task = asyncio.create_task(update_timer())
        self._timer_task.add_done_callback(done_timer)

        return

    async def stop_timer(self):
        """Cancels the timer."""

        await cancel_task(self._timer_task)
        self._timer_task = None

        if self._progress:
            self._progress.stop()
            self._progress.console.clear_live()
        self._progress = None

    def verify_files(self):
        """Checks that the files have been written and have the right contents."""

        HEADERS_CRITICAL = [
            "TILE_ID",
            "DPOS",
            "ARGON",
            "NEON",
            "LDLS",
            "QUARTZ",
            "HGNE",
            "XENON",
            "HARTMANN",
            "TESCIRA",
            "TESCIDE",
            "TESKYERA",
            "TESKYEDE",
            "TESKYWRA",
            "TESKYWDE",
            "TESPECRA",
            "TESPECDE",
        ]

        HEADERS_WARNING = []

        for file in self.get_files():
            header = fits.getheader(str(file))

            for key in HEADERS_CRITICAL:
                if key not in header:
                    raise GortSpecError(f"Keyword {key} not present in {file!s}")

            for key in HEADERS_WARNING:
                if key not in header:
                    self.specs.write_to_log(
                        f"Keyword {key} not present in {file!s}",
                        "warning",
                    )

    def get_files(self):
        """Returns the files written by the exposure."""

        sjd = get_sjd("LCO")
        config = self.specs.gort.config
        data_path = pathlib.Path(config["specs"]["data_path"].format(SJD=sjd))

        return list(data_path.glob(f"*-[0]*{self.exp_no}.fits.gz"))

    async def _done_monitor(self):
        """Waits until the spectrographs are idle, and marks the Future done."""

        await self.specs._send_command_all("wait_until_idle", allow_errored=True)

        for spec in self.specs.values():
            reply = await spec.status(simple=True)
            if "ERROR" in reply["status_names"]:
                self.error = True

        self.reading = False

        # Set the Future.
        self.set_result(self)

    async def register_observation(
        self,
        tile_id: int | None = None,
        dither_pos: int = 0,
    ):
        """Registers the exposure in the database."""

        if self.flavour != "object":
            return

        self.specs.write_to_log("Registering observation.", "info")
        registration_payload = {
            "dither": dither_pos,
            "jd": self.start_time.jd,
            "seeing": -999.0,
            "standards": [],
            "skies": [],
            "exposure_no": self.exp_no,
        }

        if tile_id is not None:
            registration_payload["tile_id"] = tile_id

        self.specs.write_to_log(f"Registration payload {registration_payload}")

        try:
            await register_observation(registration_payload)
        except Exception as err:
            self.specs.write_to_log(f"Failed registering exposure: {err}", "error")
        else:
            self.specs.write_to_log("Registration complete.")

    async def _call_hook(self, hook_name: str, *args, as_task: bool = False, **kwargs):
        """Calls the coroutines associated with a hook."""

        if hook_name not in self.hooks:
            raise ValueError(f"Invalid hook {hook_name!r}.")

        coros = self.hooks[hook_name]
        if not isinstance(coros, list) and asyncio.iscoroutinefunction(coros):
            task = asyncio.create_task(coros(*args, **kwargs))
        elif isinstance(coros, list):
            task = asyncio.gather(*[coro(*args, **kwargs) for coro in coros])
        else:
            raise ValueError(f"Invalid hook functions found for {hook_name!r}.")

        if not as_task:
            await task



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-16
# @Filename: gort.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
import pathlib
import signal
import sys
import uuid
from copy import deepcopy

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generic,
    Literal,
    Type,
    TypeVar,
)

from rich import pretty, traceback
from rich.logging import RichHandler
from typing_extensions import Self

from clu.client import AMQPClient, AMQPReply
from sdsstools.logger import SDSSLogger, get_logger

from gort import config
from gort.core import RemoteActor
from gort.exceptions import GortError
from gort.kubernetes import Kubernetes
from gort.observer import GortObserver
from gort.recipes import recipes as recipe_to_class
from gort.tile import Tile
from gort.tools import get_temporary_file_path, run_in_executor


if TYPE_CHECKING:
    from rich.console import Console

try:
    IPYTHON = get_ipython()  # type: ignore
    IPYTHON_DEFAULT_HOOKS = [
        IPYTHON._showtraceback,
        IPYTHON.showtraceback,
        IPYTHON.showsyntaxerror,
    ]
except NameError:
    IPYTHON = None
    IPYTHON_DEFAULT_HOOKS = []


__all__ = ["GortClient", "Gort", "GortDeviceSet", "GortDevice"]


DevType = TypeVar("DevType", bound="GortDeviceSet | GortDevice")


class GortClient(AMQPClient):
    """The main ``gort`` client, used to communicate with the actor system.

    A subclass of :obj:`~clu.client.AMQPClient` with defaults for host and logging,
    it loads the :obj:`.GortDeviceSet` and :obj:`.RemoteActor` instances
    for the LVM system.

    Parameters
    ----------
    host
        The host on which the RabbitMQ exchange is running.
    port
        The port on which the RabbitMQ exchange listens to connections.
    user
        The user to connect to the exchange.
    password
        The password to connect to the exchange.
    use_rich_output
        If :obj:`True`, uses ``rich`` to provide colourised tracebacks and
        prettier outputs.
    log_file_path
        The path where to save GORT's log. File logs are always saved with ``DEBUG``
        logging level. If :obj:`None`, a temporary file will be used whose path
        can be retrieved by calling :obj:`.get_log_path`. If :obj:`False`, no file
        logging will happen.

    """

    def __init__(
        self,
        host: str = "lvm-hub.lco.cl",
        port: int = 5672,
        user: str = "guest",
        password: str = "guest",
        use_rich_output: bool = True,
        log_file_path: str | pathlib.Path | Literal[False] | None = None,
    ):
        from gort.devices.ag import AGSet
        from gort.devices.enclosure import Enclosure
        from gort.devices.guider import GuiderSet
        from gort.devices.nps import NPSSet
        from gort.devices.spec import SpectrographSet
        from gort.devices.telemetry import TelemetrySet as TelemSet
        from gort.devices.telescope import TelescopeSet as TelSet

        self.client_uuid = str(uuid.uuid4()).split("-")[0]

        self.console: Console

        log = self._prepare_logger(
            log_file_path,
            use_rich_output=use_rich_output,
        )

        self._setup_exception_hooks(log, use_rich_output=use_rich_output)

        self._connect_lock = asyncio.Lock()

        super().__init__(
            f"Gort-client-{self.client_uuid}",
            host=host,
            port=port,
            user=user,
            password=password,
            log=log,
        )

        self.actors: dict[str, RemoteActor] = {}

        self.config = deepcopy(config)

        self.__device_sets = []

        self.ags = self.add_device(AGSet, self.config["ags"]["devices"])
        self.guiders = self.add_device(GuiderSet, self.config["guiders"]["devices"])
        self.telescopes = self.add_device(TelSet, self.config["telescopes"]["devices"])
        self.nps = self.add_device(NPSSet, self.config["nps"]["devices"])
        self.specs = self.add_device(SpectrographSet, self.config["specs"]["devices"])
        self.enclosure = self.add_device(Enclosure, name="enclosure", actor="lvmecp")
        self.telemetry = self.add_device(TelemSet, self.config["telemetry"]["devices"])

        try:
            self.kubernetes = Kubernetes(log=self.log)
        except Exception:
            self.log.warning(
                "Gort cannot access the Kubernets cluster. "
                "The Kubernetes module won't be available."
            )
            self.kubernetes = None

    def _prepare_logger(
        self,
        log_file_path: str | pathlib.Path | Literal[False] | None = None,
        use_rich_output: bool = True,
    ):
        """Creates a logger and start file logging."""

        log = get_logger(
            f"lvmgort-{self.client_uuid}",
            use_rich_handler=True,
            rich_handler_kwargs={"rich_tracebacks": use_rich_output},
        )

        if log_file_path is None:
            tmp_path = get_temporary_file_path(
                prefix="gort-",
                suffix=f"-{self.client_uuid}.log",
            )
            log.start_file_logger(str(tmp_path), rotating=False)
        elif log_file_path is not False:
            log.start_file_logger(str(log_file_path), rotating=False)

        assert isinstance(log.sh, RichHandler)
        self.console = log.sh.console

        return log

    def _setup_exception_hooks(self, log: SDSSLogger, use_rich_output: bool = True):
        """Setup various hooks for exception handling."""

        def custom__showtraceback_closure(default__showtraceback):
            def _showtraceback(*args, **kwargs):
                assert IPYTHON

                exc_tuple = IPYTHON._get_exc_info()
                log.handle_exceptions(*exc_tuple)

                default__showtraceback(*args, **kwargs)

            if IPYTHON:
                IPYTHON._showtraceback = _showtraceback

        if use_rich_output:
            traceback.install(console=self.console)
            pretty.install(console=self.console)

            # traceback.install() overrides the excepthook, which means that
            # tracebacks are not logged to file anymore. Restore that.
            log.default_excepthook = sys.excepthook
            sys.excepthook = log.handle_exceptions

        else:
            # Make sure we are not using rich tracebacks anymore, in
            # case we installed them at some point.
            # See https://github.com/Textualize/rich/pull/2972/files

            log.default_excepthook = sys.__excepthook__

            if IPYTHON:
                IPYTHON._showtraceback = IPYTHON_DEFAULT_HOOKS[0]
                IPYTHON.showtraceback = IPYTHON_DEFAULT_HOOKS[1]
                IPYTHON.showsyntaxerror = IPYTHON_DEFAULT_HOOKS[2]

        if IPYTHON:
            # One more override. We want that any exceptions raised in IPython
            # also gets logged to file, but IPython overrides excepthook completely
            # so here we make a custom call to log.handle_exceptions() and then
            # just let it do whatever it was its default (whether that means it
            # was overridden by rich or not).
            custom__showtraceback_closure(IPYTHON._showtraceback)

    def get_log_path(self):
        """Returns the path of the log file. :obj:`None` if not logging to file."""

        return self.log.log_filename

    async def init(self) -> Self:
        """Initialises the client.

        Returns
        -------
        object
            The same instance of :obj:`.GortClient` after initialisation.

        """

        if not self.connected:
            async with self._connect_lock:
                await self.start()

        await asyncio.gather(*[ractor.init() for ractor in self.actors.values()])

        # Initialise device sets.
        await asyncio.gather(*[dev.init() for dev in self.__device_sets])

        return self

    def add_device(self, class_: Type[DevType], *args, **kwargs) -> DevType:
        """Adds a new device or device set to Gort."""

        ds = class_(self, *args, **kwargs)
        self.__device_sets.append(ds)

        return ds

    @property
    def connected(self):
        """Returns :obj:`True` if the client is connected."""

        return self.connection and self.connection.connection is not None

    def add_actor(self, actor: str, device: GortDevice | None = None):
        """Adds an actor to the programmatic API.

        Parameters
        ----------
        actor
            The name of the actor to add.
        device
            A device associated with this actor.

        """

        if actor not in self.actors:
            self.actors[actor] = RemoteActor(self, actor, device=device)

        return self.actors[actor]

    def set_verbosity(self, verbosity: str | int | None = None):
        """Sets the level of verbosity to ``debug``, ``info``, or ``warning``.

        Parameters
        ----------
        verbosity
            The level of verbosity. Can be a string level name, an integer, or
            :obj:`None`, in which case the default verbosity will be used.

        """

        verbosity = verbosity or "warning"
        if isinstance(verbosity, int):
            verbosity = logging.getLevelName(verbosity)

        assert isinstance(verbosity, str)

        verbosity = verbosity.lower()
        if verbosity not in ["debug", "info", "warning"]:
            raise ValueError("Invalid verbosity value.")

        verbosity_level = logging.getLevelName(verbosity.upper())
        self.log.sh.setLevel(verbosity_level)


GortDeviceType = TypeVar("GortDeviceType", bound="GortDevice")


class GortDeviceSet(dict[str, GortDeviceType], Generic[GortDeviceType]):
    """A set to gort-managed devices.

    Devices can be accessed as items of the :obj:`.GortDeviceSet` dictionary
    or using dot notation, as attributes.

    Parameters
    ----------
    gort
        The :obj:`.GortClient` instance.
    data
        A mapping of device to device info. Each device must at least include
        an ``actor`` key with the actor to use to communicated with the device.
        Any other information is passed to the :obj:`.GortDevice` on instantiation.
    kwargs
        Other keyword arguments to pass wo the device class.

    """

    __DEVICE_CLASS__: ClassVar[Type["GortDevice"]]
    __DEPLOYMENTS__: ClassVar[list[str]] = []

    def __init__(self, gort: GortClient, data: dict[str, dict], **kwargs):
        self.gort = gort

        _dict_data = {}
        for device_name in data:
            device_data = data[device_name].copy()
            actor_name = device_data.pop("actor")
            _dict_data[device_name] = self.__DEVICE_CLASS__(
                gort,
                device_name,
                actor_name,
                **device_data,
                **kwargs,
            )

        dict.__init__(self, _dict_data)

    async def init(self):
        """Runs asynchronous tasks that must be executed on init."""

        # Run devices init methods.
        results = await asyncio.gather(
            *[dev.init() for dev in self.values()],
            return_exceptions=True,
        )

        for idev, result in enumerate(results):
            if isinstance(result, Exception):
                self.write_to_log(
                    f"Failed initialising device {list(self)[idev]} "
                    f"with error {str(result)}",
                    "error",
                )

        return

    def __getattribute__(self, __name: str) -> Any:
        if __name in self:
            return self.__getitem__(__name)
        return super().__getattribute__(__name)

    async def call_device_method(self, method: Callable, *args, **kwargs):
        """Calls a method in each one of the devices.

        Parameters
        ----------
        method
            The method to call. This must be the abstract class method,
            not the method from an instantiated object.
        args,kwargs
            Arguments to pass to the method.

        """

        if not callable(method):
            raise GortError("Method is not callable.")

        if hasattr(method, "__self__"):
            # This is a bound method, so let's get the class method.
            method = method.__func__

        if not hasattr(self.__DEVICE_CLASS__, method.__name__):
            raise GortError("Method does not belong to this class devices.")

        devices = self.values()

        return await asyncio.gather(*[method(dev, *args, **kwargs) for dev in devices])

    async def _send_command_all(self, command: str, *args, **kwargs):
        """Sends a command to all the devices.

        Parameters
        ----------
        command
            The command to call.
        args, kwargs
            Arguments to pass to the :obj:`.RemoteCommand`.

        """

        tasks = []
        for dev in self.values():
            actor_command = dev.actor.commands[command]
            tasks.append(actor_command(*args, **kwargs))

        return await asyncio.gather(*tasks)

    def write_to_log(
        self,
        message: str,
        level: str = "debug",
        header: str | None = None,
    ):
        """Writes a message to the log with a custom header.

        Parameters
        ----------
        message
            The message to log.
        level
            The level to use for logging: ``'debug'``, ``'info'``, ``'warning'``, or
            ``'error'``.
        header
            The header to prepend to the message. By default uses the class name.

        """

        if header is None:
            header = f"({self.__class__.__name__}) "

        message = f"{header}{message}"

        level = logging.getLevelName(level.upper())
        assert isinstance(level, int)

        self.gort.log.log(level, message)

    async def restart(self):
        """Restarts the set deployments and resets all controllers.

        Returns
        -------
        result
            A boolean indicting if the entire restart procedure succeeded.

        """

        failed: bool = False

        if self.gort.kubernetes is None:
            raise GortError("The Kubernetes cluster is not accessible.")

        self.write_to_log("Restarting Kubernetes deployments.", "info")
        for deployment in self.__DEPLOYMENTS__:
            self.gort.kubernetes.restart_deployment(deployment, from_file=True)

        self.write_to_log("Waiting 15 seconds for deployments to be ready.", "info")
        await asyncio.sleep(15)

        # Check that deployments are running.
        running_deployments = self.gort.kubernetes.list_deployments()
        for deployment in self.__DEPLOYMENTS__:
            if deployment not in running_deployments:
                failed = True
                self.write_to_log(f"Deployment {deployment} did not restart.", "error")

        # Refresh the command models for all the actors.
        await asyncio.gather(*[actor.refresh() for actor in self.gort.actors.values()])

        # Refresh the device set.
        await self.init()

        return not failed


class GortDevice:
    """A gort-managed device.

    Parameters
    ----------
    gort
        The :obj:`.GortClient` instance.
    name
        The name of the device.
    actor
        The name of the actor used to interface with this device. The actor is
        added to the list of :obj:`.RemoteActor` in the :obj:`.GortClient`.


    """

    def __init__(self, gort: GortClient, name: str, actor: str):
        self.gort = gort
        self.name = name
        self.actor = gort.add_actor(actor, device=self)

    async def init(self):
        """Runs asynchronous tasks that must be executed on init.

        If the device is part of a :obj:`.DeviceSet`, this method is called
        by :obj:`.DeviceSet.init`.

        """

        return

    def write_to_log(
        self,
        message: str,
        level: str = "debug",
        header: str | None = None,
    ):
        """Writes a message to the log with a custom header.

        Parameters
        ----------
        message
            The message to log.
        level
            The level to use for logging: ``'debug'``, ``'info'``, ``'warning'``, or
            ``'error'``.
        header
            The header to prepend to the message. By default uses the device name.

        """

        if header is None:
            header = f"({self.name}) "

        message = f"{header}{message}"

        level = logging.getLevelName(level.upper())
        assert isinstance(level, int)

        self.gort.log.log(level, message)

    def log_replies(self, reply: AMQPReply, skip_debug: bool = True):
        """Outputs command replies."""

        if reply.body:
            if reply.message_code in ["w"]:
                level = "warning"
            elif reply.message_code in ["e", "f", "!"]:
                level = "error"
            else:
                level = "debug"
                if skip_debug:
                    return

            self.write_to_log(str(reply.body), level)


class Gort(GortClient):
    """Gort's robotic functionality.

    :obj:`.Gort` is subclass of :obj:`.GortClient` that implements higher-level
    robotic functionality. This is the class a user will normally instantiate and
    interact with.

    Parameters
    ----------
    args, kwargs
        Arguments to pass to :obj:`.GortClient`.
    verbosity
        The level of logging verbosity.
    on_interrupt
        Action to perform if the loop receives an interrupt signal during
        execution. The only options are :obj:`None` (do nothing, currently
        running commands will continue), or ``stop'`` which will stop the
        telescopes and guiders before exiting.

    """

    def __init__(
        self,
        *args,
        verbosity: str | None = None,
        on_interrupt: str | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if verbosity:
            self.set_verbosity(verbosity)

        self.set_signals(mode=on_interrupt)

    def set_signals(self, mode: str | None = None):
        """Defines the behaviour when the event loop receives a signal."""

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            self.log.warning(
                "No event loop found. Signals cannot be set and "
                "this may cause other problems."
            )
            return

        async def _stop():
            await asyncio.gather(*[self.telescopes.stop(), self.guiders.stop()])
            sys.exit(1)

        if mode == "stop":
            self.log.debug(f"Adding signal handler {mode!r}.")
            loop.add_signal_handler(signal.SIGINT, lambda: asyncio.create_task(_stop()))

    async def emergency_close(self):
        """Parks and closes the telescopes."""

        tasks = []
        tasks.append(self.telescopes.park(disable=True))
        tasks.append(self.enclosure.close(force=True))

        self.log.warning("Closing and parking telescopes.")
        await asyncio.gather(*tasks)

    async def observe_tile(
        self,
        tile: Tile | int | None = None,
        ra: float | None = None,
        dec: float | None = None,
        pa: float = 0.0,
        use_scheduler: bool = False,
        exposure_time: float = 900.0,
        n_exposures: int = 1,
        guide_tolerance: float = 3.0,
        acquisition_timeout: float = 180.0,
        show_progress: bool | None = None,
    ):
        """Performs all the operations necessary to observe a tile.

        Parameters
        ----------
        tile
            The ``tile_id`` to observe, or a :obj:`.Tile` object. If not
            provided, observes the next tile suggested by the scheduler
            (requires ``use_scheduler=True``).
        ra,dec
            The RA and Dec where to point the science telescopes. The other
            telescopes are pointed to calibrators that fit the science pointing.
            Cannot be used with ``tile``.
        pa
            Position angle of the IFU. Defaults to PA=0.
        use_scheduler
            Whether to use the scheduler to determine the ``tile_id`` or
            select calibrators.
        exposure_time
            The length of the exposure in seconds.
        n_exposures
            Number of exposures to take while guiding.
        guide_tolerance
            The guide tolerance in arcsec. A telescope will not be considered
            to be guiding if its separation to the commanded field is larger
            than this value.
        acquisition_timeout
            The maximum time allowed for acquisition. In case of timeout
            the acquired fields are evaluated and an exception is
            raised if the acquisition failed.
        show_progress
            Displays a progress bar with the elapsed exposure time.

        """

        # Create tile.
        if isinstance(tile, Tile):
            pass
        elif tile is not None or (tile is None and ra is None and dec is None):
            if use_scheduler:
                tile = await run_in_executor(Tile.from_scheduler, tile_id=tile)
            else:
                raise GortError("Not enough information to create a tile.")

        elif ra is not None and dec is not None:
            if use_scheduler:
                tile = await run_in_executor(Tile.from_scheduler, ra=ra, dec=dec, pa=pa)
            else:
                tile = await run_in_executor(Tile.from_coordinates, ra, dec, pa=pa)

        else:
            raise GortError("Not enough information to create a tile.")

        assert isinstance(tile, Tile)

        # Create observer.
        observer = GortObserver(self, tile)

        # Run the cleanup routine to be extra sure.
        await self.cleanup()

        try:
            # Slew telescopes and move fibsel mask.
            await observer.slew()

            # Start guiding.
            await observer.acquire(
                guide_tolerance=guide_tolerance,
                timeout=acquisition_timeout,
            )

            # Exposing
            exposure = await observer.expose(
                exposure_time=exposure_time,
                show_progress=show_progress,
                count=n_exposures,
            )

        finally:
            # Finish observation.
            await observer.finish_observation()

        return exposure

    async def execute_recipe(self, recipe: str, **kwargs):
        """Executes a recipe.

        Parameters
        ----------
        recipe
            The name of the recipe to execute.
        kwargs
            Arguments to be passed to the recipe.

        """

        if recipe not in recipe_to_class:
            raise ValueError(f"Cannot find recipe {recipe!r}.")

        Recipe = recipe_to_class[recipe]

        return await Recipe(self)(**kwargs)

    async def startup(self, **kwargs):
        """Executes the `startup <.StartupRecipe>` sequence."""

        return await self.execute_recipe("startup", **kwargs)

    async def shutdown(self, **kwargs):
        """Executes the `shutdown <.ShutdownRecipe>` sequence."""

        return await self.execute_recipe("shutdown", **kwargs)

    async def cleanup(self, readout: bool = True):
        """Executes the `shutdown <.CleanupRecipe>` sequence."""

        return await self.execute_recipe("cleanup", readout=readout)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-13
# @Filename: operations.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING, Literal

from rich.prompt import Confirm

from .base import BaseRecipe


if TYPE_CHECKING:
    from gort.devices.spec import Spectrograph


__all__ = ["StartupRecipe", "ShutdownRecipe", "CleanupRecipe"]


OPEN_DOME_MESSAGE = """Do not open the dome if you have not checked the following:
* Humidity is below 80%
* Dew point is below the temperature by > 5 degrees (?)
* Wind is below 35 mph
* There is no-one inside the enclosure
* No rain/good conditions confirmed with the Du Pont observers

Du Pont control room:
   (US) +1 626-310-0436
   (Chile) +56 51-2203-609
Slack:
   #lvm-dupont-observing
"""


class StartupRecipe(BaseRecipe):
    """Starts the telescopes, runs the calibration sequence, and opens the enclosure."""

    name = "startup"

    async def recipe(
        self,
        calibration_sequence: str | None | Literal[False] = None,
        open_enclosure: bool = True,
        confirm_open: bool = True,
        focus: bool = True,
    ):
        """Runs the startup sequence.

        Parameters
        ----------
        gort
            The `.Gort` instance to use.
        calibration_sequence
            The name of the calibration sequence to use. If :obj:`None`, uses the
            default sequence from the configuration. If :obj:`False`, skips the
            calibration sequence.
        open_enclosure
            Whether to open the enclosure.
        confirm_open
            If :obj:`True`, asks the user to confirm opening the enclosure.
        focus
            Whether to focus after the enclosure has open.

        """

        rconfig = self.gort.config["recipes"]["startup"]

        self.gort.log.warning("Running the startup sequence.")

        await self.gort.telescopes.home(
            home_telescopes=True,
            home_kms=True,
            home_focusers=True,
            home_fibsel=True,
        )

        self.gort.log.info("Turning off all lamps.")
        await self.gort.nps.calib.all_off()

        self.gort.log.info("Taking AG darks.")
        await self.gort.guiders.take_darks()

        if calibration_sequence is not False:
            sequence = calibration_sequence or rconfig["calibration_sequence"]
            self.gort.log.info(f"Running calibration sequence {sequence!r}.")
            await self.gort.specs.calibrate(sequence)

        if open_enclosure:
            if confirm_open:
                self.gort.log.warning(OPEN_DOME_MESSAGE)
                if not Confirm.ask(
                    "Open the dome?",
                    default=False,
                    console=self.gort._console,
                ):
                    return

            self.gort.log.info("Opening the dome ...")
            await self.gort.enclosure.open()

        if open_enclosure and focus:
            self.gort.log.info("Focusing telescopes.")
            await self.gort.guiders.focus()

        self.gort.log.info("The startup recipe has completed.")


class ShutdownRecipe(BaseRecipe):
    """Closes the telescope for the night."""

    name = "shutdown"

    async def recipe(self, park_telescopes: bool = True):
        """Shutdown the telescope, closes the dome, etc.

        Parameters
        ----------
        park_telescopes
            Park telescopes (and disables axes). Set to :obj:`False` if only
            closing for a brief period of time.

        """

        self.gort.log.warning("Running the shutdown sequence.")

        self.gort.log.info("Turning off all lamps.")
        await self.gort.nps.calib.all_off()

        self.gort.log.info("Making sure guiders are idle.")
        await self.gort.guiders.stop(now=True)

        await self.gort.enclosure.close()

        if park_telescopes:
            self.gort.log.info("Parking telescopes for the night.")
            await self.gort.telescopes.park()


class CleanupRecipe(BaseRecipe):
    """Stops guiders, aborts exposures, and makes sure the system is ready to go."""

    name = "cleanup"

    async def recipe(self, readout: bool = True):
        """Runs the cleanup recipe.

        Parameters
        ----------
        readout
            If the spectrographs are idle and with a readout pending,
            reads the spectrographs.

        """

        self.gort.log.info("Stopping the guiders.")
        await self.gort.guiders.stop(now=False)

        if not (await self.gort.specs.are_idle()):
            cotasks = []

            for spec in self.gort.specs.values():
                status = await spec.status()
                names = status["status_names"]

                if await spec.is_reading():
                    self.gort.log.warning(f"{spec.name} is reading. Waiting.")
                    cotasks.append(self._wait_until_spec_is_idle(spec))
                elif await spec.is_exposing():
                    self.gort.log.warning(f"{spec.name} is exposing. Aborting.")
                    cotasks.append(spec.abort())
                elif "IDLE" in names and "READOUT_PENDING" in names:
                    msg = f"{spec.name} has a pending exposure."
                    if readout is False:
                        self.gort.log.warning(f"{msg} Aborting it.")
                        cotasks.append(spec.abort())
                    else:
                        self.gort.log.warning(f"{msg} Reading it.")
                        cotasks.append(spec.actor.commands.read())
                        cotasks.append(self._wait_until_spec_is_idle(spec))

            await asyncio.gather(*cotasks)

        self.gort.log.info("Turning all lights off.")
        await self.gort.nps.calib.all_off()

    async def _wait_until_spec_is_idle(self, spec: Spectrograph):
        """Waits until an spectrograph is idle."""

        while True:
            if await spec.is_idle():
                return

            await asyncio.sleep(3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-18
# @Filename: nps.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import datetime
from functools import partial

from typing import TYPE_CHECKING

import numpy
import pandas

from gort.exceptions import ErrorCodes, GortError, GortGuiderError
from gort.gort import GortDevice, GortDeviceSet
from gort.maskbits import GuiderStatus
from gort.tools import build_guider_reply_list, cancel_task


if TYPE_CHECKING:
    from clu import AMQPReply

    from gort.gort import GortClient


__all__ = ["Guider", "GuiderSet"]


class Guider(GortDevice):
    """Class representing a guider."""

    def __init__(self, gort: GortClient, name: str, actor: str, **kwargs):
        super().__init__(gort, name, actor)

        self.separation: float | None = None
        self.status: GuiderStatus = GuiderStatus.IDLE

        self.guide_monitor_task: asyncio.Task | None = None
        self._best_focus: tuple[float, float] = (-999.0, -999.0)

        self.gort.add_reply_callback(self._status_cb)

    @property
    def ag(self):
        """Gets the :obj:`.AG` device associated with this guider."""

        return self.gort.ags[self.name]

    @property
    def telescope(self):
        """Gets the :obj:`.Telescope` device associated with this guider."""

        return self.gort.telescopes[self.name]

    async def _status_cb(self, reply: AMQPReply):
        """Listens to guider keywords and updates the internal state."""

        if reply.sender == self.actor.name:
            if "status" in reply.body:
                self.status = GuiderStatus(int(reply.body["status"], 16))
            if "measured_pointing" in reply.body:
                self.separation = reply.body["measured_pointing"]["separation"]

    async def wait_until_guiding(
        self,
        guide_tolerance: float | None = None,
        timeout: float | None = None,
    ) -> tuple[bool, GuiderStatus, float | None, bool]:
        """Waits until the guider has converged.

        Parameters
        ----------
        guide_tolerance
            The minimum separation, in arcsec, between the measured and desired
            positions that needs to be reached before returning. If :obj:`None`,
            waits until guiding (as opposed to acquisition) begins.
        timeout
            Maximum time, in seconds, to wait before returning. If :obj:`None`,
            waits indefinitely. If the timeout is reached it does not
            raise an exception.

        Returns
        -------
        reached
            Whether the desired minimum separation was reached.
        status
            The current :obj:`.GS`.
        separation
            The current separation.
        timedout
            :obj:`True` if the acquisition timed out.

        """

        # Initial delay to allow time for the guider to switch to DRIFTING status.
        await asyncio.sleep(1)

        elapsed = 1
        while True:
            has_acquired = (
                self.status is not None
                and self.separation is not None
                and self.status & GuiderStatus.GUIDING
                and not self.status & GuiderStatus.DRIFTING
                and (guide_tolerance is None or self.separation < guide_tolerance)
            )
            if has_acquired:
                return (True, self.status, self.separation, False)

            elapsed += 1
            if timeout is not None and elapsed > timeout:
                return (False, self.status, self.separation, True)

            await asyncio.sleep(1)

    async def expose(self, *args, continuous: bool = False, **kwargs):
        """Exposes this telescope cameras.

        Parameters
        ----------
        args,kwargs
            Arguments to be passed to the guider expose command.
        continuous
            Whether to expose the camera continuously. If :obj:`False`
            it takes a single exposure.

        """

        while True:
            await self.actor.commands.expose(*args, **kwargs)

            if not continuous:
                return

    async def focus(
        self,
        inplace=False,
        guess: float | None = None,
        step_size: float = 0.5,
        steps: int = 7,
        exposure_time: float = 5.0,
    ):
        """Focus the telescope.

        Parameters
        ----------
        inplace
            If :obj:`True`, focuses the telescope where it is pointing at. Otherwise
            points to zenith.
        guess
            The initial guess for the focuser position. If :obj:`None`, the
            default value from the configuration file is used.
        step_size
            The size, in focuser units, of each step.
        steps
            The total number of step points. Must be an odd number.
        exposure_time
            The exposure time for each step.

        """

        self._best_focus = (-999, -999)

        # Send telescopes to zenith.
        if not inplace:
            self.write_to_log("Moving telescope to zenith.")
            await self.gort.telescopes[self.name].goto_named_position(
                "zenith",
                altaz_tracking=True,
            )

        if guess is None:
            guess = self.gort.config["guiders"]["focus"]["guess"][self.name]
            if isinstance(guess, (list, tuple)):
                sensor_data = await self.gort.telemetry[self.name].status()
                temp = sensor_data["sensor1"]["temperature"]
                guess = guess[0] * temp + guess[1]

        try:
            self.write_to_log(
                f"Focusing telescope {self.name} with initial guess {guess:.1f}.",
                "info",
            )
            await self.actor.commands.focus(
                reply_callback=self._parse_focus,
                guess=guess,
                step_size=step_size,
                steps=steps,
                exposure_time=exposure_time,
            )
        except GortError as err:
            self.write_to_log(f"Failed focusing with error {err}", level="error")
        finally:
            self.separation = None
            self.write_to_log(
                f"Best focus: {self._best_focus[1]} arcsec "
                f"at {self._best_focus[0]} DT",
                "info",
            )

            if self._best_focus[1] < 0.3:
                self.write_to_log("Focus value is invalid.", "error")

        return self._best_focus

    def _parse_focus(self, reply: AMQPReply):
        """Parses replies from the guider command."""

        if not reply.body:
            return

        self.log_replies(reply, skip_debug=False)

        if "best_focus" in reply.body:
            self._best_focus = (
                reply.body["best_focus"]["focus"],
                reply.body["best_focus"]["fwhm"],
            )

    async def guide(
        self,
        ra: float | None = None,
        dec: float | None = None,
        exposure_time: float = 5.0,
        pixel: tuple[float, float] | str | None = None,
        **guide_kwargs,
    ):
        """Starts the guide loop.

        This command blocks until `.stop` is called.

        Parameters
        ----------
        ra,dec
            The coordinates to acquire. If :obj:`None`, the current telescope
            coordinates are used.
        exposure_time
            The exposure time of the AG integrations.
        pixel
            The pixel on the master frame on which to guide. Defaults to
            the central pixel. This can also be the name of a known pixel
            position for this telescope, e.g., ``'P1-1'`` for ``spec``.
        guide_kwargs
            Other keyword arguments to pass to ``lvmguider guide``.

        """

        self.separation = None

        if not self.status & GuiderStatus.IDLE:
            raise GortGuiderError(
                "Guider is not IDLE",
                error_code=ErrorCodes.COMMAND_FAILED,
            )

        if ra is None or dec is None:
            status = await self.telescope.status()
            ra_status = status["ra_j2000_hours"] * 15
            dec_status = status["dec_j2000_degs"]

            ra = ra if ra is not None else ra_status
            dec = dec if dec is not None else dec_status

        config = self.gort.config
        if isinstance(pixel, str):
            if pixel not in config["guiders"]["devices"][self.name]["named_pixels"]:
                raise GortGuiderError(f"Invalid pixel name {pixel!r}.", error_code=610)
            pixel = config["guiders"]["devices"][self.name]["named_pixels"][pixel]

        log_msg = f"Guiding at RA={ra:.6f}, Dec={dec:.6f}"
        if pixel is not None:
            log_msg += f", pixel=({pixel[0]:.1f}, {pixel[1]:.1f})."
        self.write_to_log(log_msg, level="info")

        try:
            self.guide_monitor_task = asyncio.create_task(self._monitor_task())
            await self.actor.commands.guide(
                reply_callback=partial(self.log_replies, skip_debug=False),
                ra=ra,
                dec=dec,
                exposure_time=exposure_time,
                reference_pixel=pixel,
                **guide_kwargs,
            )
        finally:
            await cancel_task(self.guide_monitor_task)

    async def _monitor_task(self, timeout: float = 30):
        """Monitors guiding and reports average and last guide metrics."""

        current_data = []

        task = asyncio.create_task(
            build_guider_reply_list(
                self.gort,
                current_data,
                actor=self.actor.name,
            )
        )

        try:
            while True:
                await asyncio.sleep(timeout)

                # Build DF with all the frames.
                df = pandas.DataFrame.from_records(current_data)

                # Group by frameno, keep only non-NaN values.
                df = df.groupby(["frameno", "telescope"], as_index=False).apply(
                    lambda g: g.fillna(method="bfill", axis=0).iloc[0, :]
                )

                # Sort by frameno.
                df = df.sort_values("frameno")

                # Remove NaN rows.
                df = df.dropna()

                now = datetime.datetime.utcnow()
                time_range = now - pandas.Timedelta(f"{timeout} seconds")
                time_data = df.loc[df.time > time_range, :]

                if (
                    len(time_data) == 0
                    or "fwhm" not in time_data
                    or "separation" not in time_data
                ):
                    continue

                # Calculate and report last.
                last = time_data.tail(1)
                sep_last = round(last.separation.values[0], 3)
                fwhm_last = round(last.fwhm.values[0], 2)
                mode_last = last["mode"].values[0]
                self.write_to_log(
                    f"Last: sep={sep_last} arcsec; fwhm={fwhm_last} arcsec; "
                    f"mode={mode_last!r}",
                    "info",
                )

                # Calculate and report averages.
                sep_avg = round(time_data.separation.mean(), 3)
                fwhm_avg = round(time_data.fwhm.mean(), 2)
                self.write_to_log(
                    f"Average ({timeout} s): sep={sep_avg} arcsec; "
                    f"fwhm={fwhm_avg} arcsec",
                    "info",
                )

        except asyncio.CancelledError:
            await cancel_task(task)
            return

    async def stop(
        self,
        now: bool = False,
        wait_until_stopped: bool = True,
        timeout: float | None = 30,
    ) -> None:
        """Stops the guide loop.

        Parameters
        ----------
        now
            Aggressively stops the guide loop.
        wait_until_stopped
            Blocks until the guider is idle.
        timeout
            How long to wait for the guider to become idle before
            forcibly stop it. Ignored if ``now=True``.

        """

        self.write_to_log(f"Stopping guider with now={now}.", "info")
        await self.actor.commands.stop(now=now)

        if not now and wait_until_stopped:
            elapsed = 0.0
            while True:
                if self.status & GuiderStatus.IDLE:
                    self.write_to_log("Guider is idle.")
                    return

                await asyncio.sleep(0.5)
                elapsed += 0.5

                if timeout is not None and elapsed > timeout:
                    self.write_to_log(
                        "The guider is not yet idle. Stopping with now=True.",
                        "warning",
                    )
                    await self.stop(now=True)
                    return

        self.status = GuiderStatus.IDLE

    async def set_pixel(self, pixel: tuple[float, float] | str | None = None):
        """Sets the master frame pixel on which to guide.

        Parameters
        ----------
        pixel
            The pixel on the master frame on which to guide. Defaults to
            the central pixel. This can also be the name of a known pixel
            position for this telescope, e.g., ``'P1-1'`` for ``spec``.

        """

        config = self.gort.config
        if isinstance(pixel, str):
            if pixel not in config["guiders"]["devices"][self.name]["named_pixels"]:
                raise GortGuiderError(f"Invalid pixel name {pixel!r}.", error_code=610)
            pixel = config["guiders"]["devices"][self.name]["named_pixels"][pixel]

        if pixel is None:
            await self.actor.commands.reset_pixel()
        else:
            await self.actor.commands.set_pixel(*pixel)


class GuiderSet(GortDeviceSet[Guider]):
    """A set of telescope guiders."""

    __DEVICE_CLASS__ = Guider
    __DEPLOYMENTS__ = ["lvmguider"]

    async def expose(self, *args, continuous: bool = False, **kwargs):
        """Exposes all the cameras using the guider.

        Parameters
        ----------
        args,kwargs
            Arguments to be passed to :obj:`.Guider.expose`.
        continuous
            Whether to expose the camera continuously. If :obj:`False`
            it takes a single exposure.

        """

        await self.call_device_method(
            Guider.expose,
            *args,
            continuous=continuous,
            **kwargs,
        )

    async def take_darks(self):
        """Takes AG darks."""

        # Move telescopes to park to prevent light, since we don't have shutters.
        # We use goto_named_position to prevent disabling the telescope and having
        # to rehome.
        self.write_to_log("Moving telescopes to park position.", level="info")
        await self.gort.telescopes.goto_named_position("park")

        # Take darks.
        self.write_to_log("Taking darks.", level="info")

        cmds = []
        for guider in self.values():
            cmds.append(
                guider.actor.commands.expose(
                    flavour="dark",
                    reply_callback=partial(guider.log_replies, skip_debug=False),
                )
            )

        if len(cmds) > 0:
            await asyncio.gather(*cmds)

    async def focus(
        self,
        inplace=False,
        guess: float | dict[str, float] | None = None,
        step_size: float = 0.5,
        steps: int = 7,
        exposure_time: float = 5.0,
    ):
        """Focus all the telescopes.

        Parameters
        ----------
        inplace
            If :obj:`True`, focuses the telescopes where they are pointing at. Otherwise
            points to zenith.
        guess
            The initial guesses for focuser position. If :obj:`None`, the default
            values from the configuration file are used. It can also be a float
            value, which will be used for all telescopes, or a mapping of telescope
            name to guess value. Missing values will default to the configuration
            value.
        step_size
            The size, in focuser units, of each step.
        steps
            The total number of step points. Must be an odd number.
        exposure_time
            The exposure time for each step.

        """

        self.write_to_log("Running focus sequence.", "info")

        if guess is None:
            guess_dict = {}
        elif isinstance(guess, dict):
            guess_dict = guess
        else:
            guess_dict = {guider_name: guess for guider_name in self}

        jobs = [
            self[guider_name].focus(
                inplace=inplace,
                guess=guess_dict.get(guider_name, None),
                step_size=step_size,
                steps=steps,
                exposure_time=exposure_time,
            )
            for guider_name in self
        ]
        results = await asyncio.gather(*jobs)

        best_focus = [f"{name}: {results[itel][1]}" for itel, name in enumerate(self)]

        self.write_to_log("Best focus: " + ", ".join(best_focus), "info")

        fwhms = numpy.array([fwhm for _, fwhm in results])
        if numpy.any(fwhms < 0.3):
            self.write_to_log("One or more focus values are invalid.", "error")

    async def guide(self, *args, **kwargs):
        """Guide on all telescopes.

        Parameters
        ----------
        args,kwargs
            Arguments to be passed to :obj:`.Guider.guide`.

        """

        await self.call_device_method(Guider.guide, *args, **kwargs)

    async def stop(self, now: bool = False, wait_until_stopped: bool = True):
        """Stops the guide loop on all telescopes.

        Parameters
        ----------
        now
            Aggressively stops the guide loop.
        wait_until_stopped
            Blocks until the guider is idle.

        """

        await self.call_device_method(
            Guider.stop,
            now=now,
            wait_until_stopped=wait_until_stopped,
        )

    async def wait_until_guiding(
        self,
        names: list[str] | None = None,
        guide_tolerance: float | None = None,
        timeout: float | None = None,
    ):
        """Waits until the guiders have converged.

        Parameters
        ----------
        names
            List of telescopes to wait for convergence.
        guide_tolerance
            The minimum separation, in arcsec, between the measured and desired
            positions that needs to be reached before returning. If :obj:`None`,
            waits until guiding (as opposed to acquisition) begins.
        timeout
            Maximum time, in seconds, to wait before returning. If :obj:`None`,
            waits indefinitely. If the timeout is reached it does not
            raise an exception.

        Returns
        -------
        status
            A dictionary with the telescope names and a tuple indicating whether
            the desired minimum separation was reached. The current
            ``GuiderStatus``, and the current separation for that telescope.

        """

        names = names or list(self)
        results = await asyncio.gather(
            *[
                self[name].wait_until_guiding(
                    guide_tolerance=guide_tolerance,
                    timeout=timeout,
                )
                for name in names
            ]
        )

        return dict(zip(names, results))



# encoding: utf-8

from sdsstools import get_config, get_package_version


# pip package name
NAME = "lvmgort"

# Loads config. config name is the package name.
config = get_config(NAME, config_envvar="GORT_CONFIG_FILE")

# package name should be pip package name
__version__ = get_package_version(path=__file__, package_name=NAME)


from .core import *
from .devices import *
from .exposure import *
from .gort import *
from .observer import *
from .tile import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-01-17
# @Filename: exceptions.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import inspect
from enum import Enum

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from clu import Command

    from gort.core import RemoteCommand


def decapitalize_first_letter(s, upper_rest=False):
    return "".join([s[:1].lower(), (s[1:].upper() if upper_rest else s[1:])])


class ErrorCodes(Enum):
    """List of error codes."""

    UNCATEGORISED_ERROR = 0
    NOT_IMPLEMENTED = 1
    COMMAND_FAILED = 2
    COMMAND_TIMEDOUT = 3
    USAGE_ERROR = 4
    TIMEOUT = 5
    CANNOT_MOVE_LOCAL_MODE = 101
    FAILED_REACHING_COMMANDED_POSITION = 102
    INVALID_TELESCOPE_POSITION = 103
    FIBSEL_INVALID_POSITION = 201
    SECTROGRAPH_FAILED_EXPOSING = 301
    SECTROGRAPH_NOT_IDLE = 302
    INVALID_CALIBRATION_SEQUENCE = 303
    LOCAL_MODE_FAILED = 501
    DOOR_STATUS_FAILED = 502
    INVALID_PIXEL_NAME = 610
    ACQUISITION_FAILED = 801
    UNKNOWN_ERROR = 999


class GortError(Exception):
    """A custom core GortError exception"""

    def __init__(self, message: str | None = None, error_code: int | ErrorCodes = 0):
        try:
            self.error_code = ErrorCodes(error_code)
        except ValueError:
            self.error_code = ErrorCodes.UNKNOWN_ERROR
            error_code = self.error_code.value

        if message is not None and message != "":
            message = decapitalize_first_letter(message)
            super().__init__(f"Error {error_code} ({self.error_code.name}): {message}")
        else:
            super().__init__(f"Error {error_code} ({self.error_code.name})")


class RemoteCommandError(GortError):
    """An error in a remote command to an actor."""

    def __init__(
        self,
        message: str | None,
        command: Command,
        remote_command: RemoteCommand,
    ):
        self.command = command
        self.remote_command = remote_command
        self.actor = remote_command._remote_actor.name

        super().__init__(message, error_code=ErrorCodes.COMMAND_FAILED)


class GortTimeoutError(GortError):
    """A timeout error, potentially associated with a timed out remote command."""

    def __init__(
        self,
        message: str | None,
        command: Command | None = None,
        remote_command: RemoteCommand | None = None,
    ):
        self.command = command
        self.remote_command = remote_command
        self.actor = remote_command._remote_actor.name if remote_command else None

        if self.remote_command:
            error_code = ErrorCodes.COMMAND_TIMEDOUT
        else:
            error_code = ErrorCodes.TIMEOUT

        super().__init__(message, error_code=error_code)


class GortTimeout(GortError):
    """Raised if a timeout occurs."""

    pass


class GortNotImplemented(GortError):
    """A custom exception for not yet implemented features."""

    def __init__(self, message: str | None = None):
        message = "This feature is not implemented yet." if not message else message

        super(GortNotImplemented, self).__init__(message, error_code=1)


class GortDeviceError(GortError):
    """A device error, which appends the name of the device to the error message."""

    def __init__(
        self,
        message: str | None = None,
        error_code: int | ErrorCodes = 0,
    ) -> None:
        from gort.gort import GortDevice

        if message is not None:
            stack = inspect.stack()
            f_locals = stack[1][0].f_locals

            if "self" in f_locals:
                obj = f_locals["self"]
                name = getattr(obj, "name", None)
                if issubclass(obj.__class__, GortDevice) and name is not None:
                    message = f"({name}) {message}"

        super().__init__(message, error_code=error_code)


class GortEnclosureError(GortDeviceError):
    """Enclosure-related error."""

    pass


class GortNPSError(GortDeviceError):
    """NPS-related error."""

    pass


class GortGuiderError(GortDeviceError):
    """Guider-related error."""

    pass


class GortSpecError(GortDeviceError):
    """Spectrograph-related error."""

    pass


class GortAGError(GortDeviceError):
    """AG-related error."""

    pass


class GortTelescopeError(GortDeviceError):
    """Telescope-related error."""

    pass


class TileError(GortError):
    """An error associated with a `.Tile`."""

    pass


class GortObserverError(GortError):
    """An error associated with the `.Observer`."""

    pass


class GortWarning(Warning):
    """Base warning for Gort."""

    pass


class GortUserWarning(UserWarning, GortWarning):
    """The primary warning class."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-13
# @Filename: calibration.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import json
import pathlib
from copy import deepcopy

from typing import Any

import jsonschema

from gort.exceptions import ErrorCodes, GortSpecError
from gort.tools import cancel_task, move_mask_interval

from .base import BaseRecipe


__all__ = ["CalibrationRecipe"]


class CalibrationRecipe(BaseRecipe):
    """Runs a calibration sequence."""

    name = "calibration"

    async def recipe(
        self,
        sequence: str | dict = "normal",
        slew_telescopes: bool = True,
        park_after: bool = False,
        show_progress: bool | None = None,
    ):
        """Runs the calibration sequence.

        Parameters
        ----------
        sequence
            The name calibration sequence to execute. It can also be a
            dictionary with the calibration sequence definition that
            follows the :ref:`calibration schema <calibration-schema>`.
        slew_telescopes
            Whether to move the telescopes to point to the FF screen.
        park_after
            Park the telescopes after a successful calibration sequence.
        show_progress
            Displays a progress bar with the elapsed exposure time.

        """

        # TODO: add some checks. Confirm HDs are open, enclosure is closed,
        # specs connected, etc.

        # Calibration sequence configuration. Includes the position where to
        # point the telescopes, NPS to use, and sequences.
        cal_config = self.gort.config["recipes"]["calibration"]

        # Task that will move the fibre selector.
        fibsel_task: asyncio.Task | None = None

        sequence_config: dict[str, Any]
        if isinstance(sequence, dict):
            sequence_config = sequence

        else:
            if sequence not in cal_config["sequences"]:
                raise GortSpecError(
                    f"Unknown sequence {sequence!r}.",
                    error_code=ErrorCodes.INVALID_CALIBRATION_SEQUENCE,
                )
            sequence_config = cal_config["sequences"][sequence]

        # Validate sequence.
        schema_file = pathlib.Path(__file__).parent / "../etc/calibration_schema.json"
        schema = json.loads(open(schema_file).read())
        try:
            jsonschema.validate(sequence_config, schema)
        except jsonschema.ValidationError:
            raise GortSpecError(
                "Calibration sequence does not match schema.",
                error_code=ErrorCodes.INVALID_CALIBRATION_SEQUENCE,
            )

        calib_nps = self.gort.nps[cal_config["lamps_nps"]]

        lamps_config = sequence_config.get("lamps", {})
        has_lamps = len(lamps_config) != 0
        if not has_lamps:
            # No point in slewing if we are only taking bias and darks.
            slew_telescopes = False

        # Turn off all lamps.
        self.gort.log.info("Checking that all lamps are off.")
        await calib_nps.all_off()

        self.gort.log.info(f"Running calibration sequence {sequence!r}.")

        try:
            if "biases" in sequence_config:
                nbias = sequence_config["biases"].get("count", 1)
                self.gort.log.info(f"Taking {nbias} biases.")
                for _ in range(nbias):
                    await self.gort.specs.expose(flavour="bias", object="bias")

            if "darks" in sequence_config:
                ndarks = sequence_config["darks"].get("count", 1)
                exp_times = sequence_config["darks"]["exposure_time"]
                if isinstance(exp_times, (float, int)):
                    exp_times = [exp_times]

                self.gort.log.info(f"Taking {ndarks} x {exp_times} darks.")

                total_darks = len(exp_times) * ndarks
                idark = 1
                for exp_time in exp_times:
                    for _ in range(ndarks):
                        await self.gort.specs.expose(
                            flavour="dark",
                            exposure_time=exp_time,
                            async_readout=(idark == total_darks) and has_lamps,
                            object="dark",
                        )
                        idark += 1

            if slew_telescopes:
                # Move the telescopes to point to the screen.
                self.gort.log.info("Pointing telescopes to FF screen.")
                await self.gort.telescopes.goto_named_position(cal_config["position"])

            for lamp in lamps_config:
                warmup = lamps_config[lamp].get(
                    "warmup",
                    cal_config["defaults"]["warmup"],
                )

                self.gort.log.info(f"Warming up lamp {lamp} ({warmup} s).")
                await calib_nps.on(lamp)
                await asyncio.sleep(warmup)

                exp_times = lamps_config[lamp]["exposure_time"]
                if isinstance(exp_times, (int, float)):
                    exp_times = [exp_times]

                n_exp_times = len(exp_times)
                for ietime, exp_time in enumerate(exp_times):
                    flavour = lamps_config[lamp].get(
                        "flavour",
                        cal_config["defaults"]["flavours"][lamp.lower()],
                    )

                    # Check if we are spinning the fibre selector and,
                    # if so, launch the task.
                    fibsel = lamps_config[lamp].get("fibsel", False)
                    fibsel_def = cal_config["defaults"]["fibsel"]
                    if isinstance(fibsel, dict) or fibsel is True:
                        # If it's True, just use defaults.
                        if fibsel is True:
                            fibsel = {}

                        positions = fibsel.get("positions", fibsel_def["positions"])
                        order_by_steps = True

                        if isinstance(positions, (list, tuple)):
                            positions = list(positions)
                            order_by_steps = False
                            if "initial_position" in fibsel:
                                initial_position = fibsel["initial_position"]
                            else:
                                initial_position = positions[0]
                        else:
                            initial_position = fibsel.get(
                                "initial_position",
                                fibsel_def["initial_position"],
                            )

                        time_per_position = fibsel.get("time_per_position", None)
                        total_time = exp_time if time_per_position is None else None

                        fibsel_device = self.gort.telescopes.spec.fibsel
                        await fibsel_device.move_to_position(initial_position)

                        # Launch the task.
                        fibsel_task = asyncio.create_task(
                            move_mask_interval(
                                self.gort,
                                positions,
                                order_by_steps=order_by_steps,
                                total_time=total_time,
                                time_per_position=time_per_position,
                            )
                        )

                    self.gort.log.info(f"Exposing lamp for {exp_time} s.")
                    await self.gort.specs.expose(
                        flavour=flavour,
                        exposure_time=exp_time,
                        show_progress=show_progress,
                        async_readout=ietime == n_exp_times - 1,
                        object=lamp,
                    )

                    await cancel_task(fibsel_task)

                self.gort.log.info(f"Turning off {lamp}.")
                await calib_nps.off(lamp)

            if park_after:
                await self.gort.telescopes.park()

            if (
                self.gort.specs.last_exposure
                and not self.gort.specs.last_exposure.done()
            ):
                self.gort.log.info("Awaiting last exposure readout.")
                await self.gort.specs.last_exposure

        except Exception:
            self.gort.log.error(
                "Errored while executing sequence. "
                "Turning all the lamps off before raising.",
            )

            # Stop the mask iteration task.
            await cancel_task(fibsel_task)

            raise

        finally:
            # If there are no lamps there is no way we turned them on.
            if has_lamps:
                await calib_nps.all_off()

    def get_calibration_sequence(self, sequence: str):
        """Returns a dictionary with the configuration for a calibration sequence.

        Parameters
        ----------
        sequence
            The name calibration sequence.

        Returns
        -------
        sequence_dict
            The calibration sequence dictionary. This dictionary can be
            altered and then passed to :obj:`.calibrate` to execute the
            modified sequence. The returned dictionary if a deep copy of
            the original sequence; modifying it won't modify the original
            sequence.

        """

        sequences = self.gort.config["recipes"]["calibration"]["sequences"]

        if sequence not in sequences:
            raise ValueError(f"Sequence {sequence!r} not found in configuration file.")

        return deepcopy(sequences[sequence])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-11
# @Filename: transforms.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib
import re

import numpy
import pandas
import yaml

from gort import config


__all__ = [
    "read_fibermap",
    "offset_to_master_frame_pixel",
    "xy_to_radec_offset",
    "fibre_slew_coordinates",
    "radec_sexagesimal_to_decimal",
    "fibre_to_master_frame",
]


_FIBERMAP_CACHE: pandas.DataFrame | None = None


def read_fibermap(
    path: str | pathlib.Path | None = None,
    force_cache: bool = False,
) -> pandas.DataFrame:
    """Reads the fibermap file.

    Parameters
    ----------
    path
        Path to the fibermap file. If :obj:`None` uses the path from the
        configuration file.
    force_cache
        If :obj:`True`, forces a re-read of the fibermap file; otherwise reads it
        from the cache if available.

    Returns
    -------
    fibermap
        A pandas DataFrame with the fibermap data.

    """

    global _FIBERMAP_CACHE

    if path is None:
        path = pathlib.Path(config["lvmcore"]["path"]) / config["lvmcore"]["fibermap"]
    else:
        path = pathlib.Path(path).absolute()

    assert isinstance(path, pathlib.Path)

    if not path.exists():
        raise FileNotFoundError(f"Fibermap file {str(path)} does not exist.")

    if force_cache is False and _FIBERMAP_CACHE is not None:
        return _FIBERMAP_CACHE

    fibermap_y = yaml.load(open(str(path)), Loader=yaml.CFullLoader)

    schema = fibermap_y["schema"]
    cols = [it["name"] for it in schema]
    dtypes = [it["dtype"] if it["dtype"] != "str" else "<U8" for it in schema]

    fibers = pandas.DataFrame.from_records(
        numpy.array(
            [tuple(fibs) for fibs in fibermap_y["fibers"]],
            dtype=list(zip(cols, dtypes)),
        ),
    )

    # Lower-case some columns.
    for col in fibers:
        is_str = pandas.api.types.is_string_dtype(fibers[col].dtype)
        if is_str and col in ["targettype", "telescope"]:
            fibers[col] = fibers[col].str.lower()

    # Add a new column with the full name of the fibre, as ifulabel-finifu
    fibers["fibername"] = fibers["orig_ifulabel"]

    _FIBERMAP_CACHE = fibers

    return _FIBERMAP_CACHE


def offset_to_master_frame_pixel(
    xmm: float | None = None,
    ymm: float | None = None,
    ra: float | None = None,
    dec: float | None = None,
) -> tuple[float, float]:
    """Determines the pixel on master frame coordinates for an offset.

    Parameters
    ----------
    xmm
        The x offset, in mm, with respect to the central fibre in the IFU.
    ymm
        The y offset, in mm, with respect to the central fibre in the IFU.
    ra
        The offset in RA, in arcsec. See :obj:`xy_to_radec_offset` for the
        caveats on this calculations.
    dec
        The offset in declination, in arcsec.


    Returns
    -------
    pixel
        A tuple with the x and z coordinates of the pixel in the master frame.

    Raises
    ------
    ValueError
        If the pixel is out of bounds.

    """

    PIXEL_SIZE = 9  # microns / pixel
    PIXEL_SCALE = 1  # arcsec/pixel

    XZ_0 = (2500, 1000)  # Central pixel in the master frame

    if xmm is not None and ymm is not None:
        if ra is not None or dec is not None:
            raise ValueError("ra/dec cannot be set along with xmm/ymm.")

    elif ra is not None and dec is not None:
        if xmm is not None or ymm is not None:
            raise ValueError("xmm/ymm cannot be set along with ra/dec.")

        xmm = ra * PIXEL_SIZE / PIXEL_SCALE / 1000
        ymm = -dec * PIXEL_SIZE / PIXEL_SCALE / 1000

    else:
        raise ValueError("Not enough inputs supplied.")

    x_mf = xmm * 1000 / PIXEL_SIZE + XZ_0[0]
    y_mf = ymm * 1000 / PIXEL_SIZE + XZ_0[1]

    if x_mf < 0 or x_mf > 2 * XZ_0[0] or y_mf < 0 or y_mf > 2 * XZ_0[1]:
        raise ValueError("Pixel is out of bounds.")

    return (round(x_mf, 1), round(y_mf, 1))


def xy_to_radec_offset(xpmm: float, ypmm: float):
    """Converts offsets in the IFU to approximate RA/Dec offsets.

    .. warning::
        This is an approximate conversion that assumes the IFU is perfectly
        aligned with the AG cameras in the focal plane and that the field
        de-rotation is perfect. It should only be used to determine initial
        offsets for blind telescope slews.

    Parameters
    ----------
    xpmm
        The x offset with respect to the central IFU fibre, in mm, as defined
        in the fibermap.
    ypmm
        As ``xpmm`` for the y offset.

    Returns
    -------
    radec
        RA/Dec offset in arcsec as a tuple. See the warning above for caveats.

    """

    # In the master frame / focal plane, increase x means increasing RA and
    # increasing y means decreasing Dec (i.e., axes are rotated 180 degrees
    # wrt the usual North up, East left).

    PIXEL_SIZE = 9  # microns/pixel
    PIXEL_SCALE = 1  # arcsec/pixel

    x_arcsec = xpmm * 1000 / PIXEL_SIZE * PIXEL_SCALE
    y_arcsec = ypmm * 1000 / PIXEL_SIZE * PIXEL_SCALE

    return (x_arcsec, -y_arcsec)


def fibre_slew_coordinates(
    ra: float,
    dec: float,
    fibre_name: str,
) -> tuple[float, float]:
    """Determines the slew coordinates for a fibre.

    This function provides approximate slew coordinates to place
    a target with coordinates ``ra``, ``dec`` on a given fibre. The
    preferred way to use it is to slew the telescope using the
    coordinates returned by this function but then guide on the
    original target coordinates and use the appropriate pixel on the
    master frame to guide on the fibre.

    Parameters
    ----------
    ra,dec
        The RA and Dec of the target to which to slew.
    fibre_name
        The fibre to which to slew the target, with the format
        ``<ifulabel>-<finifu>``.

    Returns
    -------
    slew_coordinates
        A tuple of RA/Dec coordinates to slew, which should place
        the target near the desired fibre.

    """

    fibermap = read_fibermap()

    if fibre_name not in fibermap.fibername.values:
        raise NameError(f"Fibre {fibre_name} not found in fibermap.")

    fibre = fibermap.loc[fibermap.fibername == fibre_name, :]
    xpmm, ypmm = fibre.loc[:, ["xpmm", "ypmm"]].values[0]

    ra_off, dec_off = xy_to_radec_offset(xpmm, ypmm)

    ra_slew = ra + ra_off / 3600.0 / numpy.cos(numpy.radians(dec))
    dec_slew = dec + dec_off / 3600.0

    return (ra_slew, dec_slew)


def radec_sexagesimal_to_decimal(ra: str, dec: str, ra_is_hours: bool = True):
    """Converts a string of sexagesimal RA and Dec to decimal."""

    ra_match = re.match(r"([+-]?\d+?)[:hd\s](\d+)[:m\s](\d*\.?\d*)", ra)
    if ra_match is None:
        raise ValueError("Invalid format for RA.")

    ra_groups = ra_match.groups()
    ra_deg = float(ra_groups[0]) + float(ra_groups[1]) / 60 + float(ra_groups[2]) / 3600

    if ra_is_hours:
        ra_deg *= 15

    dec_match = re.match(r"([+-]?\d+?)[:hd\s](\d+)[:m\s](\d*\.?\d*)", dec)
    if dec_match is None:
        raise ValueError("Invalid format for Dec.")

    dec_groups = dec_match.groups()
    dec_deg = float(dec_groups[0])

    if dec_deg >= 0:
        dec_deg += float(dec_groups[1]) / 60 + float(dec_groups[2]) / 3600
    else:
        dec_deg -= float(dec_groups[1]) / 60 - float(dec_groups[2]) / 3600

    return ra_deg, dec_deg


def fibre_to_master_frame(fibre_name: str):
    """Returns the xz coordinates in the master frame of a named fibres.

    Parameters
    ----------
    fibre_name
        The fibre for which to calculate the master frame coordinates,
        with the format ``<ifulabel>-<finifu>``.

    Returns
    -------
    pixel
        A tuple with the x and z coordinates of the pixel in the master frame.

    Raises
    ------
    ValueError
        If the pixel is out of bounds.

    """

    XZ_0 = (2500, 1000)  # Central pixel in the master frame

    fibermap = read_fibermap()

    if fibre_name not in fibermap.fibername.values:
        raise NameError(f"Fibre {fibre_name} not found in fibermap.")

    fibre = fibermap.loc[fibermap.fibername == fibre_name, :]
    xpmm, ypmm = fibre.loc[:, ["xpmm", "ypmm"]].values[0]

    x_mf, z_mf = offset_to_master_frame_pixel(xmm=xpmm, ymm=ypmm)

    if x_mf < 0 or x_mf > 2 * XZ_0[0] or z_mf < 0 or z_mf > 2 * XZ_0[1]:
        raise ValueError("Pixel is out of bounds.")

    return (round(x_mf, 1), round(z_mf, 1))
